import asyncio
from unittest import IsolatedAsyncioTestCase
from uuid import uuid4

from ...askar.profile import AskarProfile
from ...config.injection_context import InjectionContext
from ...messaging.responder import BaseResponder
from ...tests import mock
from ...utils.testing import create_test_profile
from ...wallet.models.wallet_record import WalletRecord
from ..single_wallet_askar_manager import SingleWalletAskarMultitenantManager


class TestSingleWalletAskarMultitenantManager(IsolatedAsyncioTestCase):
    DEFAULT_MULTIENANT_WALLET_NAME = "multitenant_sub_wallet"

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.context = self.profile.context

        self.responder = mock.CoroutineMock(send=mock.CoroutineMock())
        self.context.injector.bind_instance(BaseResponder, self.responder)

        self.manager = SingleWalletAskarMultitenantManager(self.profile)

    async def test_get_wallet_profile_should_open_store_and_return_profile_with_wallet_context(
        self,
    ):
        askar_profile_mock_name = "AskarProfile"
        wallet_name = "unit_testing/" + str(uuid4())
        wallet_record = WalletRecord(
            wallet_id="test",
            settings={
                "wallet.recreate": True,
                "wallet.seed": "test_seed",
                "wallet.name": wallet_name,
                "wallet.type": "test_type",
                "wallet.rekey": "test_rekey",
                "mediation.open": True,
                "mediation.invite": "http://invite.com",
                "mediation.default_id": "24a96ef5",
                "mediation.clear": True,
            },
        )

        with (
            mock.patch(
                "acapy_agent.multitenant.single_wallet_askar_manager.wallet_config"
            ) as wallet_config,
            mock.patch(
                "acapy_agent.multitenant.single_wallet_askar_manager.AskarProfile",
            ) as AskarProfile,
        ):
            sub_wallet_profile_context = InjectionContext()
            sub_wallet_profile = AskarProfile(None, None)
            sub_wallet_profile.context.copy.return_value = sub_wallet_profile_context

            def side_effect(context, provision):
                sub_wallet_profile.name = askar_profile_mock_name
                return sub_wallet_profile, None

            wallet_config.side_effect = side_effect

            profile = await self.manager.get_wallet_profile(
                self.profile.context, wallet_record
            )

            assert profile.name == askar_profile_mock_name
            wallet_config.assert_called_once()
            wallet_config_settings_argument = wallet_config.call_args[0][0].settings
            assert (
                wallet_config_settings_argument.get("wallet.name")
                == self.DEFAULT_MULTIENANT_WALLET_NAME
            )
            assert wallet_config_settings_argument.get("wallet.id") is None
            assert wallet_config_settings_argument.get("auto_provision") is True
            assert wallet_config_settings_argument.get("wallet.type") == "askar"
            AskarProfile.assert_called_with(
                sub_wallet_profile.opened, sub_wallet_profile_context, profile_id="test"
            )
            assert sub_wallet_profile_context.settings.get("wallet.seed") == "test_seed"
            assert sub_wallet_profile_context.settings.get("wallet.rekey") == "test_rekey"
            assert sub_wallet_profile_context.settings.get("wallet.name") == wallet_name
            assert sub_wallet_profile_context.settings.get("wallet.type") == "test_type"
            assert sub_wallet_profile_context.settings.get("mediation.open") is True
            assert (
                sub_wallet_profile_context.settings.get("mediation.invite")
                == "http://invite.com"
            )
            assert (
                sub_wallet_profile_context.settings.get("mediation.default_id")
                == "24a96ef5"
            )
            assert sub_wallet_profile_context.settings.get("mediation.clear") is True
            assert (
                sub_wallet_profile_context.settings.get("wallet.id")
                == wallet_record.wallet_id
            )
            assert sub_wallet_profile_context.settings.get("wallet.name") == wallet_name
            assert (
                sub_wallet_profile_context.settings.get("wallet.askar_profile")
                == wallet_record.wallet_id
            )

    async def test_get_anoncreds_wallet_profile_should_open_store_and_return_anoncreds_profile(
        self,
    ):
        askar_profile_mock_name = "AskarProfile"
        wallet_name = "unit_testing/" + str(uuid4())
        wallet_record = WalletRecord(
            wallet_id="test",
            settings={
                "wallet.recreate": True,
                "wallet.seed": "test_seed",
                "wallet.name": wallet_name,
                "wallet.type": "askar-anoncreds",
                "wallet.rekey": "test_rekey",
            },
        )

        with (
            mock.patch(
                "acapy_agent.multitenant.single_wallet_askar_manager.wallet_config"
            ) as wallet_config,
            mock.patch(
                "acapy_agent.multitenant.single_wallet_askar_manager.AskarAnonCredsProfile",
            ) as AskarAnonCredsProfile,
        ):
            sub_wallet_profile_context = InjectionContext()
            sub_wallet_profile = AskarAnonCredsProfile(None, None)
            sub_wallet_profile.context.copy.return_value = sub_wallet_profile_context

            def side_effect(context, provision):
                sub_wallet_profile.name = askar_profile_mock_name
                return sub_wallet_profile, None

            wallet_config.side_effect = side_effect

            await self.manager.get_wallet_profile(self.profile.context, wallet_record)

            AskarAnonCredsProfile.assert_called_with(
                sub_wallet_profile.opened, sub_wallet_profile_context, profile_id="test"
            )

    async def test_get_wallet_profile_should_create_profile(self):
        wallet_record = WalletRecord(wallet_id="test", settings={})
        create_profile_stub = asyncio.Future()
        create_profile_stub.set_result("")

        with mock.patch(
            "acapy_agent.multitenant.single_wallet_askar_manager.AskarProfile"
        ) as AskarProfile:
            sub_wallet_profile = AskarProfile(None, None)
            sub_wallet_profile.context.copy.return_value = InjectionContext()
            sub_wallet_profile.store.create_profile.return_value = create_profile_stub
            self.manager._multitenant_profile = sub_wallet_profile

            await self.manager.get_wallet_profile(
                self.profile.context, wallet_record, provision=True
            )

            sub_wallet_profile.store.create_profile.assert_called_once_with(
                wallet_record.wallet_id
            )

    async def test_get_wallet_profile_should_use_custom_subwallet_name(self):
        wallet_record = WalletRecord(wallet_id="test", settings={})
        multitenant_sub_wallet_name = "custom_wallet_name"
        self.profile.context.settings = self.profile.settings.extend(
            {"multitenant.wallet_name": multitenant_sub_wallet_name}
        )

        with mock.patch(
            "acapy_agent.multitenant.single_wallet_askar_manager.wallet_config"
        ) as wallet_config:
            with mock.patch(
                "acapy_agent.multitenant.single_wallet_askar_manager.AskarProfile"
            ) as AskarProfile:
                sub_wallet_profile = AskarProfile(None, None)
                sub_wallet_profile.context.copy.return_value = InjectionContext()

                def side_effect(context, provision):
                    return sub_wallet_profile, None

                wallet_config.side_effect = side_effect

                await self.manager.get_wallet_profile(self.profile.context, wallet_record)

                wallet_config.assert_called_once()
                assert (
                    wallet_config.call_args[0][0].settings.get("wallet.name")
                    == multitenant_sub_wallet_name
                )

    async def test_remove_wallet_profile(self):
        test_profile = await create_test_profile({"wallet.id": "test"})

        with mock.patch.object(AskarProfile, "remove") as profile_remove:
            await self.manager.remove_wallet_profile(test_profile)
            profile_remove.assert_called_once_with()

    async def test_open_profiles(self):
        assert len(list(self.manager.open_profiles)) == 0

        create_profile_stub = asyncio.Future()
        create_profile_stub.set_result("")
        with mock.patch(
            "acapy_agent.multitenant.single_wallet_askar_manager.AskarProfile"
        ) as AskarProfile:
            sub_wallet_profile = AskarProfile(None, None)
            sub_wallet_profile.context.copy.return_value = InjectionContext()
            sub_wallet_profile.store.create_profile.return_value = create_profile_stub
            self.manager._multitenant_profile = sub_wallet_profile

        assert len(list(self.manager.open_profiles)) == 1



"""Common endorsement utilities."""

import logging

from ..connections.models.conn_record import ConnRecord
from ..core.profile import Profile
from ..protocols.endorse_transaction.v1_0.manager import TransactionManager
from ..protocols.endorse_transaction.v1_0.transaction_jobs import TransactionJob
from ..protocols.endorse_transaction.v1_0.util import (
    get_endorser_connection_id,
    is_author_role,
)
from ..protocols.out_of_band.v1_0.manager import OutOfBandManager
from ..protocols.out_of_band.v1_0.messages.invitation import InvitationMessage

LOGGER = logging.getLogger(__name__)


class EndorsementSetupError(Exception):
    """Endorsement setup error."""


async def attempt_auto_author_with_endorser_setup(profile: Profile):
    """Automatically setup the author's endorser connection if possible."""

    if not is_author_role(profile):
        return

    endorser_alias = profile.settings.get_value("endorser.endorser_alias")
    if not endorser_alias:
        LOGGER.info("No endorser alias, alias is required if invitation is specified.")
        return

    connection_id = await get_endorser_connection_id(profile)
    if connection_id:
        LOGGER.info("Connected to endorser from previous connection.")
        return

    endorser_did = profile.settings.get_value("endorser.endorser_public_did")
    if not endorser_did:
        LOGGER.info("No endorser DID, can connect, but can't setup connection metadata.")
        return

    endorser_invitation = profile.settings.get_value("endorser.endorser_invitation")
    if not endorser_invitation:
        LOGGER.info("No endorser invitation, can't create connection automatically.")
        return

    try:
        # OK, we are an author, we have no endorser connection but we have enough info
        # to automatically initiate the connection
        invite = InvitationMessage.from_url(endorser_invitation)
        if invite:
            oob_mgr = OutOfBandManager(profile)
            oob_record = await oob_mgr.receive_invitation(
                invitation=invite,
                auto_accept=True,
                alias=endorser_alias,
            )
            async with profile.session() as session:
                conn_record = await ConnRecord.retrieve_by_id(
                    session, oob_record.connection_id
                )
        else:
            raise EndorsementSetupError("Invalid OOB Invitation URL")

        # configure the connection role and info (don't need to wait for the connection)
        transaction_mgr = TransactionManager(profile)
        await transaction_mgr.set_transaction_my_job(
            record=conn_record,
            transaction_my_job=TransactionJob.TRANSACTION_AUTHOR.name,
        )

        async with profile.session() as session:
            value = await conn_record.metadata_get(session, "endorser_info")
            if value:
                value["endorser_did"] = endorser_did
                value["endorser_name"] = endorser_alias
            else:
                value = {"endorser_did": endorser_did, "endorser_name": endorser_alias}
            await conn_record.metadata_set(session, key="endorser_info", value=value)

        LOGGER.info(
            "Successfully connected to endorser from invitation, and setup connection metadata."  # noqa: E501
        )

    except Exception:
        LOGGER.info("Error accepting endorser invitation/configuring endorser connection")



"""Transformation between StorageRecord and VCRecord."""

import json

from ...storage.record import StorageRecord
from .vc_record import VCRecord

VC_CRED_RECORD_TYPE = "vc_cred"


def storage_to_vc_record(record: StorageRecord) -> VCRecord:
    """Convert an Indy-SDK stored record into a VC record."""
    cred_tags = {}
    contexts = []
    types = []
    schema_ids = []
    subject_ids = []
    proof_types = []
    issuer_id = None
    given_id = None
    for tagname, tagval in (record.tags or {}).items():
        if tagname.startswith("ctxt:"):
            contexts.append(tagname[5:])
        elif tagname.startswith("type:"):
            types.append(tagname[5:])
        elif tagname.startswith("schm:"):
            schema_ids.append(tagname[5:])
        elif tagname.startswith("subj:"):
            subject_ids.append(tagname[5:])
        elif tagname.startswith("ptyp:"):
            proof_types.append(tagname[5:])
        elif tagname == "issuer_id":
            issuer_id = tagval
        elif tagname == "given_id":
            given_id = tagval
        else:
            cred_tags[tagname] = tagval
    return VCRecord(
        contexts=contexts,
        expanded_types=types,
        schema_ids=schema_ids,
        issuer_id=issuer_id,
        subject_ids=subject_ids,
        proof_types=proof_types,
        cred_value=json.loads(record.value),
        given_id=given_id,
        cred_tags=cred_tags,
        record_id=record.id,
    )


def vc_to_storage_record(cred: VCRecord) -> StorageRecord:
    """Convert a VC record into an in-memory stored record."""
    tags = {}
    for ctx_val in cred.contexts:
        tags[f"ctxt:{ctx_val}"] = "1"
    for type_val in cred.expanded_types:
        tags[f"type:{type_val}"] = "1"
    for schema_val in cred.schema_ids:
        tags[f"schm:{schema_val}"] = "1"
    for subj_id in cred.subject_ids:
        tags[f"subj:{subj_id}"] = "1"
    for proof_type in cred.proof_types:
        tags[f"ptyp:{proof_type}"] = "1"
    if cred.issuer_id:
        tags["issuer_id"] = cred.issuer_id
    if cred.given_id:
        tags["given_id"] = cred.given_id
    if cred.cred_tags:
        tags.update(cred.cred_tags)

    return StorageRecord(
        VC_CRED_RECORD_TYPE,
        json.dumps(cred.cred_value),
        tags,
        cred.record_id,
    )



"""Represents the attached message to be included in the transaction record."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import ATTACHED_MESSAGE

SCHEMA_TYPE = "101"
PROTOCOL_VERSION = "2"


class MessagesAttach(AgentMessage):
    """Class representing the attached message."""

    class Meta:
        """Metadata for attached message class."""

        schema_class = "MessagesAttachSchema"
        message_type = ATTACHED_MESSAGE

    def __init__(
        self,
        *,
        author_did: Optional[str] = None,
        author_verkey: Optional[str] = None,
        endorser_did: Optional[str] = None,
        transaction_message: Optional[dict] = None,
        transaction_type: Optional[str] = None,
        mechanism: Optional[str] = None,
        taaDigest: Optional[str] = None,
        time: Optional[int] = None,
        **kwargs,
    ):
        """Initialize the attached message object.

        Args:
            author_did: The public did of the author who creates the transaction
            author_verkey: The verkey of the author who creates the transaction
            endorser_did: The public did of the endorser who endorses the transaction
            transaction_message: The actual data present in the transaction payload
            transaction_type: The type of transaction
            mechanism: The mechanism of the latest TAA present on the ledger
            taaDigest: The digest of the latest TAA present on the ledger
            time: The time when the latest TAA was set/enabled
            kwargs: Additional keyword arguments for the message
        """

        super().__init__(**kwargs)

        transaction_message = transaction_message or {}
        self.mime_type = "application/json"

        self.data = {
            "json": {
                "endorser": endorser_did,
                "identifier": author_did,
                "operation": {
                    "data": transaction_message,
                    "type": transaction_type,
                },
                "protocol_version": PROTOCOL_VERSION,
                "reqId": 1597766666168851000,
                "signatures": {author_did: author_verkey},
                "taaAcceptance": {
                    "mechanism": mechanism,
                    "taaDigest": taaDigest,
                    "time": time,
                },
            }
        }


class MessagesAttachSchema(AgentMessageSchema):
    """Attached Message schema class."""

    class Meta:
        """Attached message schema metadata."""

        model_class = MessagesAttach
        unknown = EXCLUDE

    mime_type = fields.Str(required=True, metadata={"example": "application/json"})
    data = fields.Dict(
        required=True,
        metadata={
            "example": {
                "json": {
                    "endorser": "V4SGRU86Z58d6TV7PBUe6f",
                    "identifier": "LjgpST2rjsoxYegQDRm7EL",
                    "operation": {
                        "data": {
                            "attr_names": ["first_name", "last_name"],
                            "name": "test_schema",
                            "version": "2.1",
                        },
                        "type": "101",
                    },
                    "protocolVersion": 2,
                    "reqId": 1597766666168851000,
                    "signatures": {
                        "LjgpST2rjs": (
                            "4uq1mUATWKZArwyuPgjUEw5UBysWNbkf2SN6SqVwbfSqCfnbm1Vnfw"
                        )
                    },
                    "taaAcceptance": {
                        "mechanism": "manual",
                        "taaDigest": (
                            "f50feca7bd4202c2ab977006761d36bd6f23e4c6a7e0fc2feb9f62"
                        ),
                        "time": 1597708800,
                    },
                }
            }
        },
    )



from unittest import IsolatedAsyncioTestCase

from .....admin.request_context import AdminRequestContext
from .....storage.error import StorageError, StorageNotFoundError
from .....tests import mock
from .....utils.testing import create_test_profile
from .....wallet.did_method import DIDMethods
from .. import routes as test_module
from ..models.mediation_record import MediationRecord
from ..route_manager import RouteManager


class TestCoordinateMediationRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())
        self.context = AdminRequestContext.test_context({}, profile=self.profile)
        self.outbound_message_router = mock.CoroutineMock()
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": self.outbound_message_router,
        }
        self.request = mock.MagicMock(
            match_info={
                "mediation_id": "test-mediation-id",
                "conn_id": "test-conn-id",
            },
            query={},
            json=mock.CoroutineMock(return_value={}),
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )
        serialized = {
            "mediation_id": "fake_id",
            "state": "granted",
            "role": "server",
            "connection_id": "c3dd00cf-f6a2-4ddf-93d8-49ae74bdacef",
            "routing_keys": ["EwUKjVLboiLSuoWSEtDvrgrd41EUxG5bLecQrkHB63Up"],
            "endpoint": "http://192.168.1.13:3005",
            "created_at": "1234567890",
        }
        self.mock_record = mock.MagicMock(
            **serialized,
            serialize=mock.MagicMock(return_value=serialized),
            save=mock.CoroutineMock(),
        )

    def test_mediation_sort_key(self):
        assert (
            test_module.mediation_sort_key(
                {"state": MediationRecord.STATE_DENIED, "created_at": ""}
            )
            == "2"
        )
        assert (
            test_module.mediation_sort_key(
                {"state": MediationRecord.STATE_REQUEST, "created_at": ""}
            )
            == "1"
        )
        assert (
            test_module.mediation_sort_key(
                {"state": MediationRecord.STATE_GRANTED, "created_at": ""}
            )
            == "0"
        )

    async def test_list_mediation_requests(self):
        self.request.query = {}
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "query",
                mock.CoroutineMock(return_value=[self.mock_record]),
            ) as mock_query,
            mock.patch.object(test_module.web, "json_response") as json_response,
        ):
            await test_module.list_mediation_requests(self.request)
            json_response.assert_called_once_with(
                {"results": [self.mock_record.serialize.return_value]}
            )
            assert mock_query.mock_calls[0][2] == {}

    async def test_list_mediation_requests_filters(self):
        self.request.query = {
            "state": MediationRecord.STATE_GRANTED,
            "conn_id": "test-conn-id",
        }
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "query",
                mock.CoroutineMock(return_value=[self.mock_record]),
            ) as mock_query,
            mock.patch.object(test_module.web, "json_response") as json_response,
        ):
            await test_module.list_mediation_requests(self.request)
            json_response.assert_called_once_with(
                {"results": [self.mock_record.serialize.return_value]}
            )
            assert mock_query.return_value[0].state == MediationRecord.STATE_GRANTED

    async def test_list_mediation_requests_x(self):
        with mock.patch.object(
            test_module,
            "MediationRecord",
            mock.MagicMock(
                query=mock.CoroutineMock(side_effect=test_module.StorageError())
            ),
        ):
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.list_mediation_requests(self.request)

    async def test_list_mediation_requests_no_records(self):
        with (
            mock.patch.object(
                test_module,
                "MediationRecord",
                mock.MagicMock(query=mock.CoroutineMock(return_value=[])),
            ),
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            await test_module.list_mediation_requests(self.request)
            mock_response.assert_called_once_with({"results": []})

    async def test_retrieve_mediation_request(self):
        with (
            mock.patch.object(
                test_module.MediationRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_mediation_record_retrieve,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_mediation_record_retrieve.return_value = self.mock_record
            await test_module.retrieve_mediation_request(self.request)
            mock_response.assert_called_once_with(self.mock_record.serialize.return_value)
            mock_mediation_record_retrieve.assert_called()

    async def test_retrieve_mediation_request_x_not_found(self):
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
            ),
            mock.patch.object(test_module.web, "json_response"),
            self.assertRaises(test_module.web.HTTPNotFound),
        ):
            await test_module.retrieve_mediation_request(self.request)

    async def test_retrieve_mediation_request_x_storage_error(self):
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageError()),
            ),
            mock.patch.object(test_module.web, "json_response"),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.retrieve_mediation_request(self.request)

    async def test_delete_mediation_request(self):
        with (
            mock.patch.object(
                test_module.MediationRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_mediation_record_retrieve,
            mock.patch.object(
                self.mock_record, "delete_record", mock.CoroutineMock()
            ) as mock_delete_record,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_mediation_record_retrieve.return_value = self.mock_record
            await test_module.delete_mediation_request(self.request)
            mock_response.assert_called_once_with(self.mock_record.serialize.return_value)
            mock_mediation_record_retrieve.assert_called()
            mock_delete_record.assert_called()

    async def test_delete_mediation_request_x_not_found(self):
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
            ),
            mock.patch.object(test_module.web, "json_response"),
            self.assertRaises(test_module.web.HTTPNotFound),
        ):
            await test_module.delete_mediation_request(self.request)

    async def test_delete_mediation_request_x_storage_error(self):
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageError()),
            ),
            mock.patch.object(test_module.web, "json_response"),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.delete_mediation_request(self.request)

    async def test_request_mediation(self):
        body = {}
        self.request.json.return_value = body
        with (
            mock.patch.object(
                test_module, "MediationManager", autospec=True
            ) as mock_med_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(
                test_module.MediationRecord,
                "exists_for_connection_id",
                mock.CoroutineMock(return_value=False),
            ),
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ),
        ):
            mock_med_mgr.return_value.prepare_request = mock.CoroutineMock(
                return_value=(
                    self.mock_record,
                    mock.MagicMock(  # mediation request
                        serialize=mock.MagicMock(return_value={"a": "value"}),
                    ),
                )
            )
            await test_module.request_mediation(self.request)
            mock_response.assert_called_once_with(
                self.mock_record.serialize.return_value, status=201
            )
            self.outbound_message_router.assert_called()

    async def test_request_mediation_x_conn_not_ready(self):
        body = {}
        self.request.json.return_value = body
        with (
            mock.patch.object(
                test_module.ConnRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=mock.MagicMock(is_ready=False)),
            ),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.request_mediation(self.request)

    async def test_request_mediation_x_already_exists(self):
        body = {}
        self.request.json.return_value = body
        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ),
            mock.patch.object(
                test_module.MediationRecord,
                "exists_for_connection_id",
                mock.CoroutineMock(return_value=True),
            ),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.request_mediation(self.request)

    async def test_request_mediation_x_conn_not_found(self):
        body = {}
        self.request.json.return_value = body
        with (
            mock.patch.object(
                test_module.ConnRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
            ),
            self.assertRaises(test_module.web.HTTPNotFound),
        ):
            await test_module.request_mediation(self.request)

    async def test_request_mediation_x_storage_error(self):
        body = {}
        self.request.json.return_value = body
        with (
            mock.patch.object(
                test_module.ConnRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageError()),
            ),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.request_mediation(self.request)

    async def test_mediation_request_grant_role_server(self):
        self.mock_record.role = MediationRecord.ROLE_SERVER
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=self.mock_record),
            ),
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            await test_module.mediation_request_grant(self.request)
            mock_response.assert_called_once_with(
                self.mock_record.serialize.return_value, status=201
            )
            self.outbound_message_router.assert_called()

    async def test_mediation_request_grant_role_client_x(self):
        self.mock_record.role = MediationRecord.ROLE_CLIENT
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=self.mock_record),
            ),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.mediation_request_grant(self.request)

    async def test_mediation_request_grant_x_rec_not_found(self):
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
            ),
            self.assertRaises(test_module.web.HTTPNotFound),
        ):
            await test_module.mediation_request_grant(self.request)

    async def test_mediation_request_grant_x_storage_error(self):
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageError()),
            ),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.mediation_request_grant(self.request)

    async def test_mediation_request_deny_role_server(self):
        self.mock_record.role = MediationRecord.ROLE_SERVER
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=self.mock_record),
            ),
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            await test_module.mediation_request_deny(self.request)
            mock_response.assert_called_once_with(
                self.mock_record.serialize.return_value, status=201
            )
            self.outbound_message_router.assert_called()

    async def test_mediation_request_deny_role_client_x(self):
        self.mock_record.role = MediationRecord.ROLE_CLIENT
        with (
            mock.patch.object(
                test_module.MediationRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_mediation_record_retrieve,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_mediation_record_retrieve.return_value = mock.MagicMock(
                role=MediationRecord.ROLE_CLIENT
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.mediation_request_deny(self.request)

    async def test_mediation_request_deny_x_rec_not_found(self):
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
            ),
            self.assertRaises(test_module.web.HTTPNotFound),
        ):
            await test_module.mediation_request_deny(self.request)

    async def test_mediation_request_deny_x_storage_error(self):
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageError()),
            ),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.mediation_request_deny(self.request)

    async def test_get_keylist(self):
        session = await self.profile.session()
        self.request.query["role"] = MediationRecord.ROLE_SERVER
        self.request.query["conn_id"] = "test-id"

        query_results = [
            mock.MagicMock(
                serialize=mock.MagicMock(return_value={"serialized": "route record"})
            )
        ]

        with (
            mock.patch.object(
                test_module.RouteRecord,
                "query",
                mock.CoroutineMock(return_value=query_results),
            ) as mock_query,
            mock.patch.object(
                self.profile,
                "session",
                mock.MagicMock(return_value=session),
            ) as mock_session,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            await test_module.get_keylist(self.request)
            mock_response.assert_called_once_with(
                {"results": [{"serialized": "route record"}]}, status=200
            )
            mock_query.assert_called_once_with(
                mock_session.return_value,
                {"connection_id": "test-id", "role": MediationRecord.ROLE_SERVER},
            )

    async def test_get_keylist_no_matching_records(self):
        session = await self.profile.session()
        with (
            mock.patch.object(
                test_module.RouteRecord,
                "query",
                mock.CoroutineMock(return_value=[]),
            ) as mock_query,
            mock.patch.object(
                self.profile,
                "session",
                mock.MagicMock(return_value=session),
            ) as mock_session,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            await test_module.get_keylist(self.request)
            mock_query.assert_called_once_with(mock_session.return_value, {})
            mock_response.assert_called_once_with({"results": []}, status=200)

    async def test_get_keylist_storage_error(self):
        with (
            mock.patch.object(
                test_module.RouteRecord,
                "query",
                mock.CoroutineMock(side_effect=test_module.StorageError),
            ),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.get_keylist(self.request)

    async def test_send_keylist_update(self):
        body = {
            "updates": [
                {
                    "recipient_key": "EwUKjVLboiLSuoWSEtDvrgrd41EUxG5bLecQrkHB63Up",
                    "action": "add",
                },
                {
                    "recipient_key": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
                    "action": "remove",
                },
            ]
        }
        body_with_didkey = {
            "updates": [
                {
                    "recipient_key": "did:key:z6MktPjNKjb39Fpv2JM8vTBmhnQcsaWLN9Kx2fXLh2FC1GGC",
                    "action": "add",
                },
                {
                    "recipient_key": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                    "action": "remove",
                },
            ]
        }

        self.request.json.return_value = body

        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        state=MediationRecord.STATE_GRANTED, connection_id="test-conn-id"
                    )
                ),
            ),
            mock.patch.object(
                test_module.web,
                "json_response",
                mock.MagicMock(
                    side_effect=lambda *args, **kwargs: [*args, *kwargs.values()]
                ),
            ),
        ):
            results, status = await test_module.send_keylist_update(self.request)
            assert results["updates"] == body_with_didkey["updates"]
            assert status == 201

    async def test_send_keylist_update_bad_action(self):
        self.request.json.return_value = {
            "updates": [
                {
                    "recipient_key": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
                    "action": "wrong",
                },
            ]
        }

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.send_keylist_update(self.request)

    async def test_send_keylist_update_bad_mediation_state(self):
        self.request.json.return_value = {
            "updates": [
                {
                    "recipient_key": "EwUKjVLboiLSuoWSEtDvrgrd41EUxG5bLecQrkHB63Up",
                    "action": "add",
                },
            ]
        }

        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        state=MediationRecord.STATE_DENIED, connection_id="test-conn-id"
                    )
                ),
            ),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.send_keylist_update(self.request)

    async def test_send_keylist_update_bad_updates(self):
        self.request.json.return_value = {"updates": []}
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.send_keylist_update(self.request)

    async def test_send_keylist_update_x_no_mediation_rec(self):
        self.request.json.return_value = {
            "updates": [
                {
                    "recipient_key": "EwUKjVLboiLSuoWSEtDvrgrd41EUxG5bLecQrkHB63Up",
                    "action": "add",
                },
            ]
        }
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
            ),
            self.assertRaises(test_module.web.HTTPNotFound),
        ):
            await test_module.send_keylist_update(self.request)

    async def test_send_keylist_update_x_storage_error(self):
        self.request.json.return_value = {
            "updates": [
                {
                    "recipient_key": "EwUKjVLboiLSuoWSEtDvrgrd41EUxG5bLecQrkHB63Up",
                    "action": "add",
                },
            ]
        }

        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageError()),
            ),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.send_keylist_update(self.request)

    @mock.patch.object(test_module, "MediationManager", autospec=True)
    async def test_send_keylist_query(self, mock_manager):
        self.request.json.return_value = {"filter": {"test": "filter"}}
        self.request.query = {"paginate_limit": 10, "paginate_offset": 20}
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=self.mock_record),
            ),
            mock.patch.object(
                mock_manager.return_value,
                "prepare_keylist_query",
                mock.CoroutineMock(),
            ) as mock_prepare_keylist_query,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            await test_module.send_keylist_query(self.request)
            mock_prepare_keylist_query.assert_called_once_with(
                filter_={"test": "filter"}, paginate_limit=10, paginate_offset=20
            )
            self.outbound_message_router.assert_called()
            mock_response.assert_called_once_with(
                mock_prepare_keylist_query.return_value.serialize.return_value,
                status=201,
            )

    async def test_send_keylist_query_x_no_mediation_record(self):
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
            ),
            self.assertRaises(test_module.web.HTTPNotFound),
        ):
            await test_module.send_keylist_query(self.request)

    async def test_send_keylist_query_x_storage_error(self):
        with (
            mock.patch.object(
                test_module.MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageError()),
            ),
            self.assertRaises(test_module.web.HTTPBadRequest),
        ):
            await test_module.send_keylist_query(self.request)

    async def test_get_default_mediator(self):
        self.request.query = {}
        with (
            mock.patch.object(test_module.web, "json_response") as json_response,
            mock.patch.object(
                test_module.MediationManager,
                "get_default_mediator",
                mock.CoroutineMock(return_value=self.mock_record),
            ),
        ):
            await test_module.get_default_mediator(self.request)
            json_response.assert_called_once_with(
                self.mock_record.serialize.return_value,
                status=200,
            )

    async def test_get_empty_default_mediator(self):
        self.request.query = {}
        with (
            mock.patch.object(test_module.web, "json_response") as json_response,
            mock.patch.object(
                test_module.MediationManager,
                "get_default_mediator",
                mock.CoroutineMock(return_value=None),
            ),
        ):
            await test_module.get_default_mediator(self.request)
            json_response.assert_called_once_with(
                {},
                status=200,
            )

    async def test_get_default_mediator_storage_error(self):
        self.request.query = {}
        with (
            mock.patch.object(test_module.web, "json_response"),
            mock.patch.object(
                test_module.MediationManager,
                "get_default_mediator",
                mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
            ),
        ):
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.get_default_mediator(self.request)

    async def test_set_default_mediator(self):
        self.request.match_info = {
            "mediation_id": "fake_id",
        }
        self.request.query = {}
        with (
            mock.patch.object(
                test_module.MediationManager,
                "get_default_mediator",
                mock.CoroutineMock(return_value=self.mock_record),
            ),
            mock.patch.object(
                test_module.MediationManager,
                "set_default_mediator_by_id",
                mock.CoroutineMock(),
            ),
            mock.patch.object(test_module.web, "json_response") as json_response,
        ):
            await test_module.set_default_mediator(self.request)
            json_response.assert_called_once_with(
                self.mock_record.serialize.return_value,
                status=201,
            )

    async def test_set_default_mediator_storage_error(self):
        self.request.match_info = {
            "mediation_id": "bad_id",
        }
        self.request.query = {}
        with (
            mock.patch.object(
                test_module.MediationManager,
                "get_default_mediator",
                mock.CoroutineMock(side_effect=test_module.StorageError()),
            ),
            mock.patch.object(
                test_module.MediationManager,
                "set_default_mediator_by_id",
                mock.CoroutineMock(side_effect=test_module.StorageError()),
            ),
            mock.patch.object(test_module.web, "json_response"),
        ):
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.set_default_mediator(self.request)

    async def test_clear_default_mediator(self):
        self.request.query = {}
        with (
            mock.patch.object(
                test_module.MediationManager,
                "get_default_mediator",
                mock.CoroutineMock(return_value=self.mock_record),
            ),
            mock.patch.object(
                test_module.MediationManager,
                "clear_default_mediator",
                mock.CoroutineMock(),
            ),
            mock.patch.object(test_module.web, "json_response") as json_response,
        ):
            await test_module.clear_default_mediator(self.request)
            json_response.assert_called_once_with(
                self.mock_record.serialize.return_value,
                status=201,
            )

    async def test_clear_default_mediator_storage_error(self):
        self.request.query = {}
        with (
            mock.patch.object(
                test_module.MediationManager,
                "get_default_mediator",
                mock.CoroutineMock(side_effect=test_module.StorageError()),
            ),
            mock.patch.object(
                test_module.MediationManager,
                "clear_default_mediator",
                mock.CoroutineMock(),
            ),
            mock.patch.object(test_module.web, "json_response"),
        ):
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.clear_default_mediator(self.request)

    async def test_update_keylist_for_connection(self):
        self.request.query = {}
        self.request.json.return_value = {"mediation_id": "test-mediation-id"}
        self.request.match_info = {
            "conn_id": "test-conn-id",
        }
        mock_route_manager = mock.MagicMock(RouteManager)
        mock_keylist_update = mock.MagicMock()
        mock_keylist_update.serialize.return_value = {"mock": "serialized"}
        mock_route_manager.route_connection = mock.CoroutineMock(
            return_value=mock_keylist_update
        )
        mock_route_manager.mediation_record_for_connection = mock.CoroutineMock()
        self.context.injector.bind_instance(RouteManager, mock_route_manager)
        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ),
            mock.patch.object(test_module.web, "json_response") as json_response,
        ):
            await test_module.update_keylist_for_connection(self.request)
            json_response.assert_called_once_with({"mock": "serialized"}, status=200)

    async def test_update_keylist_for_connection_not_found(self):
        self.request.query = {}
        self.request.json.return_value = {"mediation_id": "test-mediation-id"}
        self.request.match_info = {
            "conn_id": "test-conn-id",
        }
        mock_route_manager = mock.MagicMock(RouteManager)
        mock_keylist_update = mock.MagicMock()
        mock_keylist_update.serialize.return_value = {"mock": "serialized"}
        mock_route_manager.route_connection = mock.CoroutineMock(
            return_value=mock_keylist_update
        )
        mock_route_manager.mediation_record_for_connection = mock.CoroutineMock()
        self.context.injector.bind_instance(RouteManager, mock_route_manager)
        with mock.patch.object(
            test_module.ConnRecord,
            "retrieve_by_id",
            mock.CoroutineMock(side_effect=StorageNotFoundError),
        ):
            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.update_keylist_for_connection(self.request)

    async def test_update_keylist_for_connection_storage_error(self):
        self.request.query = {}
        self.request.json.return_value = {"mediation_id": "test-mediation-id"}
        self.request.match_info = {
            "conn_id": "test-conn-id",
        }
        mock_route_manager = mock.MagicMock(RouteManager)
        mock_keylist_update = mock.MagicMock()
        mock_keylist_update.serialize.return_value = {"mock": "serialized"}
        mock_route_manager.route_connection = mock.CoroutineMock(
            return_value=mock_keylist_update
        )
        mock_route_manager.mediation_record_for_connection = mock.CoroutineMock()
        self.context.injector.bind_instance(RouteManager, mock_route_manager)
        with mock.patch.object(
            test_module.ConnRecord,
            "retrieve_by_id",
            mock.CoroutineMock(side_effect=StorageError),
        ):
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.update_keylist_for_connection(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



"""Handler for incoming disclose messages."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ..manager import V20DiscoveryMgr
from ..messages.disclosures import Disclosures


class DisclosuresHandler(BaseHandler):
    """Handler for incoming disclose messages."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler implementation."""
        self._logger.debug("DiscloseHandler called with context %s", context)
        assert isinstance(context.message, Disclosures)

        if not context.connection_ready:
            raise HandlerException(
                "Received disclosures message from inactive connection"
            )

        profile = context.profile
        mgr = V20DiscoveryMgr(profile)
        await mgr.receive_disclose(
            context.message, connection_id=context.connection_record.connection_id
        )



"""Recover a revocation registry."""

import hashlib
import importlib
import logging
import tempfile
import time

import aiohttp
import base58

LOGGER = logging.getLogger(__name__)


"""
This module calculates a new ledger accumulator, based on the revocation status
on the ledger vs revocations recorded in the wallet.
The calculated transaction can be written to the ledger to get the ledger back
in sync with the wallet.
This function can be used if there were previous revocation errors (i.e. the
credential revocation was successfully written to the wallet but the ledger write
failed.)
"""


class RevocRecoveryException(Exception):
    """Raise exception generating the recovery transaction."""


async def fetch_txns(genesis_txns, registry_id):
    """Fetch tails file and revocation registry information."""

    try:
        vdr_module = importlib.import_module("indy_vdr")
        credx_module = importlib.import_module("indy_credx")
    except Exception as e:
        raise RevocRecoveryException(f"Failed to import library {e}")

    pool = await vdr_module.open_pool(transactions=genesis_txns)
    LOGGER.debug("Connected to pool")

    LOGGER.debug("Fetch registry: %s", registry_id)
    fetch = vdr_module.ledger.build_get_revoc_reg_def_request(None, registry_id)
    result = await pool.submit_request(fetch)
    if not result["data"]:
        raise RevocRecoveryException(f"Registry definition not found for {registry_id}")
    data = result["data"]
    data["ver"] = "1.0"
    defn = credx_module.RevocationRegistryDefinition.load(data)
    LOGGER.debug("Tails URL: %s", defn.tails_location)

    async with aiohttp.ClientSession() as session:
        data = await session.get(defn.tails_location)
        tails_data = await data.read()
        tails_hash = base58.b58encode(hashlib.sha256(tails_data).digest()).decode("utf-8")
        if tails_hash != defn.tails_hash:
            raise RevocRecoveryException(
                f"Tails hash mismatch {tails_hash} {defn.tails_hash}"
            )
        else:
            LOGGER.debug("Checked tails hash: %s", tails_hash)
        tails_temp = tempfile.NamedTemporaryFile(delete=False)
        tails_temp.write(tails_data)
        tails_temp.close()

    to_timestamp = int(time.time())
    fetch = vdr_module.ledger.build_get_revoc_reg_delta_request(
        None, registry_id, None, to_timestamp
    )
    result = await pool.submit_request(fetch)
    if not result["data"]:
        raise RevocRecoveryException("Error fetching delta from ledger")

    accum_to = result["data"]["value"]["accum_to"]
    accum_to["ver"] = "1.0"
    delta = credx_module.RevocationRegistryDelta.load(accum_to)
    registry = credx_module.RevocationRegistry.load(accum_to)
    LOGGER.debug("Ledger registry state: %s", registry.to_json())
    revoked = set(result["data"]["value"]["revoked"])
    LOGGER.debug("Ledger revoked indexes: %s", revoked)

    return defn, registry, delta, revoked, tails_temp


async def generate_ledger_rrrecovery_txn(
    genesis_txns, registry_id, set_revoked, cred_def, rev_reg_def_private
):
    """Generate a new ledger accum entry, based on wallet vs ledger revocation state."""

    new_delta = None

    ledger_data = await fetch_txns(genesis_txns, registry_id)
    if not ledger_data:
        return new_delta
    defn, registry, delta, prev_revoked, tails_temp = ledger_data

    set_revoked = set(set_revoked)
    mismatch = prev_revoked - set_revoked
    if mismatch:
        LOGGER.warning(
            "Credential index(es) revoked on the ledger, but not in wallet: %s",
            mismatch,
        )

    updates = set_revoked - prev_revoked
    if not updates:
        LOGGER.debug("No updates to perform")
    else:
        LOGGER.debug("New revoked indexes: %s", updates)

        LOGGER.debug("tails_temp: %s", tails_temp.name)
        update_registry = registry.copy()
        new_delta = update_registry.update(
            cred_def, defn, rev_reg_def_private, [], updates
        )

        LOGGER.debug("New delta:")
        LOGGER.debug(new_delta.to_json())

    return new_delta



from unittest import IsolatedAsyncioTestCase

from ...connections.models.conn_record import ConnRecord
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import endorsement_setup
from ..endorsement_setup import attempt_auto_author_with_endorser_setup

mock_invitation = "http://localhost:9030?oob=eyJAdHlwZSI6ICJodHRwczovL2RpZGNvbW0ub3JnL291dC1vZi1iYW5kLzEuMS9pbnZpdGF0aW9uIiwgIkBpZCI6ICI2MWU1MmYzZS1kNTliLTQ3OWYtYmYwNC04NjJlOTk1MmM4MDYiLCAibGFiZWwiOiAiZW5kb3JzZXIiLCAiaGFuZHNoYWtlX3Byb3RvY29scyI6IFsiaHR0cHM6Ly9kaWRjb21tLm9yZy9kaWRleGNoYW5nZS8xLjAiXSwgInNlcnZpY2VzIjogW3siaWQiOiAiI2lubGluZSIsICJ0eXBlIjogImRpZC1jb21tdW5pY2F0aW9uIiwgInJlY2lwaWVudEtleXMiOiBbImRpZDprZXk6ejZNa2VkRDMyZlZmOG5ReG5SS2QzUmQ5S1hZQnVETEJiOHUyM1JWMm1ReFlpanR2I3o2TWtlZEQzMmZWZjhuUXhuUktkM1JkOUtYWUJ1RExCYjh1MjNSVjJtUXhZaWp0diJdLCAic2VydmljZUVuZHBvaW50IjogImh0dHA6Ly9sb2NhbGhvc3Q6OTAzMCJ9XX0="


class MockConnRecord:
    connection_id = "test-connection-id"


class TestEndorsementSetupUtil(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile()

    @mock.patch.object(endorsement_setup.LOGGER, "info", return_value=mock.MagicMock())
    async def test_not_enough_configs_for_connection(self, mock_logger):
        await endorsement_setup.attempt_auto_author_with_endorser_setup(self.profile)

        # No invitation
        self.profile.settings.set_value("endorser.author", True)
        await endorsement_setup.attempt_auto_author_with_endorser_setup(self.profile)

        # No endorser alias
        self.profile.settings.set_value("endorser.endorser_invitation", mock_invitation)
        await endorsement_setup.attempt_auto_author_with_endorser_setup(self.profile)

        # No endorser DID
        self.profile.settings.set_value("endorser.endorser_alias", "test-alias")
        await endorsement_setup.attempt_auto_author_with_endorser_setup(self.profile)

        assert mock_logger.call_count == 3
        for call in mock_logger.call_args_list:
            assert "Error accepting endorser invitation" not in call[0][0]

    @mock.patch.object(endorsement_setup.LOGGER, "info", return_value=mock.MagicMock())
    @mock.patch.object(endorsement_setup, "OutOfBandManager")
    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=ConnRecord(connection_id="test-connection-id"),
    )
    async def test_create_connection_with_valid_invitation(
        self, mock_conn_record, mock_oob_manager, mock_logger
    ):
        mock_oob_manager.return_value.receive_invitation = mock.CoroutineMock(
            return_value=MockConnRecord()
        )
        self.profile.settings.set_value("endorser.author", True)
        self.profile.settings.set_value("endorser.endorser_invitation", mock_invitation)
        self.profile.settings.set_value("endorser.endorser_alias", "test-alias")
        self.profile.settings.set_value("endorser.endorser_public_did", "test-did")

        await attempt_auto_author_with_endorser_setup(self.profile)

        for call in mock_logger.call_args_list:
            assert "Error accepting endorser invitation" not in call[0][0]

        assert mock_conn_record.called

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_alias",
        return_value=[ConnRecord(connection_id="test-connection-id")],
    )
    @mock.patch.object(endorsement_setup.LOGGER, "info", return_value=mock.MagicMock())
    async def test_has_previous_connection(self, mock_logger, *_):
        await endorsement_setup.attempt_auto_author_with_endorser_setup(self.profile)

        # No invitation
        self.profile.settings.set_value("endorser.author", True)
        self.profile.settings.set_value("endorser.endorser_alias", "test-alias")
        await endorsement_setup.attempt_auto_author_with_endorser_setup(self.profile)

        assert mock_logger.call_count == 1
        assert (
            mock_logger.call_args_list[0][0][0]
            == "Connected to endorser from previous connection."
        )

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_alias",
        side_effect=Exception("No connection"),
    )
    @mock.patch.object(endorsement_setup.LOGGER, "info", return_value=mock.MagicMock())
    async def test_does_not_have_previous_connection_with_did_but_no_invitation(
        self, mock_logger, *_
    ):
        await endorsement_setup.attempt_auto_author_with_endorser_setup(self.profile)

        # No invitation
        self.profile.settings.set_value("endorser.author", True)
        self.profile.settings.set_value("endorser.endorser_alias", "test-alias")
        self.profile.settings.set_value(
            "endorser.endorser_public_did", "WuE7ndRJgAGbJYnwDXV7Pz"
        )
        await endorsement_setup.attempt_auto_author_with_endorser_setup(self.profile)



"""DIDComm v1 envelope handling via Askar backend."""

from collections import OrderedDict
from typing import Optional, Sequence, Tuple

from aries_askar import Key, KeyAlg, Session, crypto_box
from aries_askar.bindings import key_get_secret_bytes
from marshmallow import ValidationError

from ...utils.jwe import JweEnvelope, JweRecipient, b64url
from ...wallet.base import WalletError
from ...wallet.crypto import extract_pack_recipients
from ...wallet.util import b58_to_bytes, bytes_to_b58


def pack_message(
    to_verkeys: Sequence[str], from_key: Optional[Key], message: bytes
) -> bytes:
    """Encode a message using the DIDComm v1 'pack' algorithm."""
    wrapper = JweEnvelope(with_protected_recipients=True, with_flatten_recipients=False)
    cek = Key.generate(KeyAlg.C20P)
    # avoid converting to bytes object: this way the only copy is zeroed afterward
    cek_b = key_get_secret_bytes(cek._handle)
    sender_vk = (
        bytes_to_b58(from_key.get_public_bytes()).encode("utf-8") if from_key else None
    )
    sender_xk = from_key.convert_key(KeyAlg.X25519) if from_key else None

    for target_vk in to_verkeys:
        target_xk = Key.from_public_bytes(
            KeyAlg.ED25519, b58_to_bytes(target_vk)
        ).convert_key(KeyAlg.X25519)
        if sender_vk:
            enc_sender = crypto_box.crypto_box_seal(target_xk, sender_vk)
            nonce = crypto_box.random_nonce()
            enc_cek = crypto_box.crypto_box(target_xk, sender_xk, cek_b, nonce)
            wrapper.add_recipient(
                JweRecipient(
                    encrypted_key=enc_cek,
                    header=OrderedDict(
                        [
                            ("kid", target_vk),
                            ("sender", b64url(enc_sender)),
                            ("iv", b64url(nonce)),
                        ]
                    ),
                )
            )
        else:
            enc_sender = None
            nonce = None
            enc_cek = crypto_box.crypto_box_seal(target_xk, cek_b)
            wrapper.add_recipient(
                JweRecipient(encrypted_key=enc_cek, header={"kid": target_vk})
            )
    wrapper.set_protected(
        OrderedDict(
            [
                ("enc", "xchacha20poly1305_ietf"),
                ("typ", "JWM/1.0"),
                ("alg", "Authcrypt" if from_key else "Anoncrypt"),
            ]
        ),
    )
    enc = cek.aead_encrypt(message, aad=wrapper.protected_bytes)
    ciphertext, tag, nonce = enc.parts
    wrapper.set_payload(ciphertext, nonce, tag)
    return wrapper.to_json().encode("utf-8")


async def unpack_message(session: Session, enc_message: bytes) -> Tuple[str, str, str]:
    """Decode a message using the DIDComm v1 'unpack' algorithm."""
    try:
        wrapper = JweEnvelope.from_json(enc_message)
    except ValidationError:
        raise WalletError("Invalid packed message")

    alg = wrapper.protected.get("alg")
    is_authcrypt = alg == "Authcrypt"
    if not is_authcrypt and alg != "Anoncrypt":
        raise WalletError("Unsupported pack algorithm: {}".format(alg))

    recips = extract_pack_recipients(wrapper.recipients)

    payload_key, sender_vk = None, None
    for recip_vk in recips:
        recip_key_entry = await session.fetch_key(recip_vk)
        if recip_key_entry:
            payload_key, sender_vk = _extract_payload_key(
                recips[recip_vk], recip_key_entry.key
            )
            break

    if not payload_key:
        raise WalletError(
            "No corresponding recipient key found in {}".format(tuple(recips))
        )
    if not sender_vk and is_authcrypt:
        raise WalletError("Sender public key not provided for Authcrypt message")

    cek = Key.from_secret_bytes(KeyAlg.C20P, payload_key)
    message = cek.aead_decrypt(
        wrapper.ciphertext,
        nonce=wrapper.iv,
        tag=wrapper.tag,
        aad=wrapper.protected_bytes,
    )
    return message, recip_vk, sender_vk


def _extract_payload_key(sender_cek: dict, recip_secret: Key) -> Tuple[bytes, str]:
    """Extract the payload key from pack recipient details.

    Returns: A tuple of the CEK and sender verkey
    """
    recip_x = recip_secret.convert_key(KeyAlg.X25519)

    if sender_cek["nonce"] and sender_cek["sender"]:
        sender_vk = crypto_box.crypto_box_seal_open(recip_x, sender_cek["sender"]).decode(
            "utf-8"
        )
        sender_x = Key.from_public_bytes(
            KeyAlg.ED25519, b58_to_bytes(sender_vk)
        ).convert_key(KeyAlg.X25519)
        cek = crypto_box.crypto_box_open(
            recip_x, sender_x, sender_cek["key"], sender_cek["nonce"]
        )
    else:
        sender_vk = None
        cek = crypto_box.crypto_box_seal_open(recip_x, sender_cek["key"])
    return cek, sender_vk



"""Utilities to deal with indy."""

from typing import Optional

from ...indy.holder import IndyHolder
from .pres_preview import IndyPresPreview


async def indy_proof_req_preview2indy_requested_creds(
    indy_proof_req: dict,
    preview: Optional[IndyPresPreview] = None,
    *,
    holder: IndyHolder,
):
    """Build indy requested-credentials structure.

    Given input proof request and presentation preview, use credentials in
    holder's wallet to build indy requested credentials structure for input
    to proof creation.

    Args:
        indy_proof_req: indy proof request
        preview: preview from presentation proposal, if applicable
        holder: holder injected into current context

    """
    req_creds = {
        "self_attested_attributes": {},
        "requested_attributes": {},
        "requested_predicates": {},
    }

    for referent, req_item in indy_proof_req["requested_attributes"].items():
        credentials = await holder.get_credentials_for_presentation_request_by_referent(
            presentation_request=indy_proof_req,
            referents=(referent,),
            offset=0,
            limit=100,
        )
        if not credentials:
            raise ValueError(
                "Could not automatically construct presentation for "
                + f"presentation request {indy_proof_req['name']}"
                + f":{indy_proof_req['version']} because referent "
                + f"{referent} did not produce any credentials."
            )

        # match returned creds against any preview values
        if len(credentials) == 1:
            cred_match = credentials[0]
        elif preview:
            reft = indy_proof_req["requested_attributes"][referent]
            names = [reft["name"]] if "name" in reft else reft.get("names")
            for cred in credentials:  # holder sorts by irrevocability, least referent
                if all(
                    preview.has_attr_spec(
                        cred_def_id=cred["cred_info"]["cred_def_id"],
                        name=name,
                        value=cred["cred_info"]["attrs"][name],
                    )
                    for name in names
                ):
                    cred_match = cred
                    break
            else:
                raise ValueError(
                    "Could not automatically construct presentation for "
                    + f"presentation request {indy_proof_req['name']}"
                    + f":{indy_proof_req['version']} because referent "
                    + f"{referent} did not produce any credentials matching "
                    + "proposed preview."
                )
        else:
            cred_match = credentials[0]  # holder sorts

        if "restrictions" in indy_proof_req["requested_attributes"][referent]:
            req_creds["requested_attributes"][referent] = {
                "cred_id": cred_match["cred_info"]["referent"],
                "revealed": True,
            }
        else:
            req_creds["self_attested_attributes"][referent] = cred_match["cred_info"][
                "attrs"
            ][indy_proof_req["requested_attributes"][referent]["name"]]

    for referent in indy_proof_req["requested_predicates"]:
        credentials = await holder.get_credentials_for_presentation_request_by_referent(
            presentation_request=indy_proof_req,
            referents=(referent,),
            offset=0,
            limit=100,
        )
        if not credentials:
            raise ValueError(
                "Could not automatically construct presentation for "
                + f"presentation request {indy_proof_req['name']}"
                + f":{indy_proof_req['version']} because predicate "
                + f"referent {referent} did not produce any credentials."
            )

        cred_match = credentials[0]  # holder sorts
        if "restrictions" in indy_proof_req["requested_predicates"][referent]:
            req_creds["requested_predicates"][referent] = {
                "cred_id": cred_match["cred_info"]["referent"],
                "revealed": True,
            }
        else:
            req_creds["self_attested_attributes"][referent] = cred_match["cred_info"][
                "attrs"
            ][indy_proof_req["requested_predicates"][referent]["name"]]

    return req_creds


def indy_proof_req2non_revoc_intervals(indy_proof_req: dict):
    """Return non-revocation intervals by requested item referent in proof request."""
    non_revoc_intervals = {}
    for req_item_type in ("requested_attributes", "requested_predicates"):
        for reft, req_item in indy_proof_req[req_item_type].items():
            interval = req_item.get(
                "non_revoked",
                indy_proof_req.get("non_revoked"),
            )
            if interval:
                timestamp_from = interval.get("from")
                timestamp_to = interval.get("to")
                if (timestamp_to is not None) and timestamp_from == timestamp_to:
                    interval["from"] = 0  # accommodate indy-sdk verify=False if fro=to
            non_revoc_intervals[reft] = interval
    return non_revoc_intervals



"""did method.py contains registry for did methods."""

from enum import Enum
from typing import Dict, List, Mapping, Optional

from .error import BaseError
from .key_type import BLS12381G2, ED25519, P256, X25519, KeyType


class HolderDefinedDid(Enum):
    """Define if a holder can specify its own did for a given method."""

    NO = "no"  # holder CANNOT provide a DID
    ALLOWED = "allowed"  # holder CAN provide a DID
    REQUIRED = "required"  # holder MUST provide a DID


class DIDMethod:
    """Class to represent a did method."""

    def __init__(
        self,
        name: str,
        key_types: List[KeyType],
        rotation: bool = False,
        holder_defined_did: HolderDefinedDid = HolderDefinedDid.NO,
    ):
        """Construct did method class."""
        self._method_name: str = name
        self._supported_key_types: List[KeyType] = key_types
        self._supports_rotation: bool = rotation
        self._holder_defined_did: HolderDefinedDid = holder_defined_did

    @property
    def method_name(self):
        """Get method name."""
        return self._method_name

    @property
    def supports_rotation(self):
        """Check rotation support."""
        return self._supports_rotation

    @property
    def supported_key_types(self):
        """Get supported key types."""
        return self._supported_key_types

    def supports_key_type(self, key_type: KeyType) -> bool:
        """Check whether the current method supports the key type."""
        return key_type in self.supported_key_types

    def holder_defined_did(self) -> HolderDefinedDid:
        """Return the did derivation policy.

        eg: did:key DIDs are derived from the verkey -> HolderDefinedDid.NO
        eg: did:web DIDs cannot be derived from key material -> HolderDefinedDid.REQUIRED
        """
        return self._holder_defined_did


SOV = DIDMethod(
    name="sov",
    key_types=[ED25519],
    rotation=True,
    holder_defined_did=HolderDefinedDid.ALLOWED,
)
INDY = DIDMethod(
    name="indy",
    key_types=[ED25519],
    rotation=True,
    holder_defined_did=HolderDefinedDid.ALLOWED,
)
KEY = DIDMethod(
    name="key",
    key_types=[ED25519, P256, BLS12381G2],
    rotation=False,
)
WEB = DIDMethod(
    name="web",
    key_types=[ED25519, BLS12381G2],
    rotation=True,
    holder_defined_did=HolderDefinedDid.REQUIRED,
)
PEER2 = DIDMethod(
    name="did:peer:2",
    key_types=[ED25519, X25519],
    rotation=False,
    holder_defined_did=HolderDefinedDid.NO,
)

PEER4 = DIDMethod(
    name="did:peer:4",
    key_types=[ED25519, X25519],
    rotation=False,
    holder_defined_did=HolderDefinedDid.NO,
)

WEBVH = DIDMethod(
    name="webvh",
    key_types=[ED25519, X25519],
    rotation=False,
    holder_defined_did=HolderDefinedDid.NO,
)


class DIDMethods:
    """DID Method class specifying DID methods with supported key types."""

    def __init__(self) -> None:
        """Construct did method registry."""
        self._registry: Dict[str, DIDMethod] = {
            SOV.method_name: SOV,
            INDY.method_name: INDY,
            KEY.method_name: KEY,
            WEB.method_name: WEB,
            PEER2.method_name: PEER2,
            PEER4.method_name: PEER4,
            WEBVH.method_name: WEBVH,
        }

    def registered(self, method: str) -> bool:
        """Check for a supported method."""
        return method in self._registry.keys()

    def register(self, method: DIDMethod):
        """Register a new did method."""
        self._registry[method.method_name] = method

    def from_method(self, method_name: str) -> Optional[DIDMethod]:
        """Retrieve a did method from method name."""
        return self._registry.get(method_name)

    def from_metadata(self, metadata: Mapping) -> Optional[DIDMethod]:
        """Get DID method instance from metadata object.

        Returns SOV if no metadata was found for backwards compatibility.
        """
        method_name: str = metadata.get("method", "sov")
        return self.from_method(method_name)

    def from_did(self, did: str) -> DIDMethod:
        """Get DID method instance from the did url."""
        method_name = did.split(":")[1] if did.startswith("did:") else SOV.method_name
        method: DIDMethod | None = self.from_method(method_name)
        if not method:
            raise BaseError(f"Unsupported did method: {method_name}")
        return method



EXAMPLES_V1 = {
    "@context": [
        {"@version": 1.1},
        "https://www.w3.org/ns/odrl.jsonld",
        {
            "ex": "https://example.org/examples#",
            "schema": "http://schema.org/",
            "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "3rdPartyCorrelation": "ex:3rdPartyCorrelation",
            "AllVerifiers": "ex:AllVerifiers",
            "Archival": "ex:Archival",
            "BachelorDegree": "ex:BachelorDegree",
            "Child": "ex:Child",
            "CLCredentialDefinition2019": "ex:CLCredentialDefinition2019",
            "CLSignature2019": "ex:CLSignature2019",
            "IssuerPolicy": "ex:IssuerPolicy",
            "HolderPolicy": "ex:HolderPolicy",
            "Mother": "ex:Mother",
            "RelationshipCredential": "ex:RelationshipCredential",
            "UniversityDegreeCredential": "ex:UniversityDegreeCredential",
            "ZkpExampleSchema2018": "ex:ZkpExampleSchema2018",
            "issuerData": "ex:issuerData",
            "attributes": "ex:attributes",
            "signature": "ex:signature",
            "signatureCorrectnessProof": "ex:signatureCorrectnessProof",
            "primaryProof": "ex:primaryProof",
            "nonRevocationProof": "ex:nonRevocationProof",
            "alumniOf": {"@id": "schema:alumniOf", "@type": "rdf:HTML"},
            "child": {"@id": "ex:child", "@type": "@id"},
            "degree": "ex:degree",
            "degreeType": "ex:degreeType",
            "degreeSchool": "ex:degreeSchool",
            "college": "ex:college",
            "name": {"@id": "schema:name", "@type": "rdf:HTML"},
            "givenName": "schema:givenName",
            "familyName": "schema:familyName",
            "parent": {"@id": "ex:parent", "@type": "@id"},
            "referenceId": "ex:referenceId",
            "documentPresence": "ex:documentPresence",
            "evidenceDocument": "ex:evidenceDocument",
            "spouse": "schema:spouse",
            "subjectPresence": "ex:subjectPresence",
            "verifier": {"@id": "ex:verifier", "@type": "@id"},
        },
    ]
}



from unittest import IsolatedAsyncioTestCase

from aiohttp import web

from acapy_agent.admin.request_context import AdminRequestContext
from acapy_agent.did.indy.indy_manager import DidIndyManager
from acapy_agent.did.indy.routes import create_indy_did
from acapy_agent.tests import mock
from acapy_agent.utils.testing import create_test_profile
from acapy_agent.wallet.did_method import DIDMethods
from acapy_agent.wallet.error import WalletError


class TestDidIndyRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            },
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

    @mock.patch.object(
        DidIndyManager,
        "register",
        return_value={"did": "did:indy:DFZgMggBEXcZFVQ2ZBTwdr", "verkey": "BnSWTUQmdYC"},
    )
    async def test_create_indy_did(self, mock_register):
        self.profile.context.injector.bind_instance(
            DIDMethods, mock.MagicMock(DIDMethods, auto_spec=True)
        )
        self.request.json = mock.CoroutineMock(return_value={})
        response = await create_indy_did(self.request)
        assert response.status == 200
        assert mock_register.called

        self.request.json = mock.CoroutineMock(
            return_value={
                "features": {},
                "options": {
                    "did": "did:indy:WRfXPg8dantKVubE3HX8pw",
                    "key_type": "ed25519",
                },
            }
        )
        response = await create_indy_did(self.request)
        assert response.status == 200
        assert mock_register.called

    @mock.patch.object(
        DidIndyManager,
        "register",
        side_effect=[WalletError("Error creating DID")],
    )
    async def test_create_indy_did_wallet_error(self, _):
        self.profile.context.injector.bind_instance(
            DIDMethods, mock.MagicMock(DIDMethods, auto_spec=True)
        )
        self.request.json = mock.CoroutineMock(return_value={})
        with self.assertRaises(web.HTTPBadRequest):
            await create_indy_did(self.request)



"""Issuer revocation registry storage handling."""

import importlib
import json
import logging
from functools import total_ordering
from os.path import join
from pathlib import Path
from shutil import move
from typing import Any, Mapping, Optional, Sequence, Tuple, Union
from urllib.parse import urlparse

from marshmallow import fields, validate
from uuid_utils import uuid4

from ...core.profile import Profile, ProfileSession
from ...indy.credx.issuer import (
    CATEGORY_CRED_DEF,
    CATEGORY_REV_REG,
    CATEGORY_REV_REG_DEF_PRIVATE,
)
from ...indy.issuer import IndyIssuer, IndyIssuerError
from ...indy.models.revocation import (
    IndyRevRegDef,
    IndyRevRegDefSchema,
    IndyRevRegEntry,
    IndyRevRegEntrySchema,
)
from ...indy.util import indy_client_dir
from ...ledger.base import BaseLedger
from ...ledger.error import LedgerError, LedgerTransactionError
from ...messaging.models.base_record import BaseRecord, BaseRecordSchema
from ...messaging.valid import (
    BASE58_SHA256_HASH_EXAMPLE,
    BASE58_SHA256_HASH_VALIDATE,
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_DID_EXAMPLE,
    INDY_DID_VALIDATE,
    INDY_REV_REG_ID_EXAMPLE,
    INDY_REV_REG_ID_VALIDATE,
    UUID4_EXAMPLE,
)
from ...tails.indy_tails_server import IndyTailsServer
from ..error import RevocationError
from ..recover import generate_ledger_rrrecovery_txn
from .issuer_cred_rev_record import IssuerCredRevRecord
from .revocation_registry import RevocationRegistry

DEFAULT_REGISTRY_SIZE = 1000

LOGGER = logging.getLogger(__name__)


@total_ordering
class IssuerRevRegRecord(BaseRecord):
    """Class for managing local issuing revocation registries."""

    class Meta:
        """IssuerRevRegRecord metadata."""

        schema_class = "IssuerRevRegRecordSchema"

    RECORD_ID_NAME = "record_id"
    RECORD_TYPE = "issuer_rev_reg"
    RECORD_TOPIC = "revocation_registry"
    LOG_STATE_FLAG = "debug.revocation"
    TAG_NAMES = {
        "cred_def_id",
        "issuer_did",
        "revoc_def_type",
        "revoc_reg_id",
        "state",
    }

    REVOC_DEF_TYPE_CL = "CL_ACCUM"

    STATE_INIT = "init"
    STATE_GENERATED = "generated"
    STATE_POSTED = "posted"  # definition published
    STATE_ACTIVE = "active"  # initial entry published, possibly subsequent entries
    STATE_FULL = "full"  # includes corrupt
    STATE_DECOMMISSIONED = "decommissioned"

    STATES = [
        STATE_INIT,
        STATE_GENERATED,
        STATE_POSTED,
        STATE_ACTIVE,
        STATE_FULL,
        STATE_DECOMMISSIONED,
    ]
    TERMINAL_STATES = [STATE_FULL, STATE_DECOMMISSIONED]

    def __init__(
        self,
        *,
        record_id: Optional[str] = None,
        state: Optional[str] = None,
        cred_def_id: Optional[str] = None,
        error_msg: Optional[str] = None,
        issuer_did: Optional[str] = None,
        max_cred_num: Optional[int] = None,
        revoc_def_type: Optional[str] = None,
        revoc_reg_id: Optional[str] = None,
        revoc_reg_def: Union[IndyRevRegDef, Mapping] = None,
        revoc_reg_entry: Union[IndyRevRegEntry, Mapping] = None,
        tag: Optional[str] = None,
        tails_hash: Optional[str] = None,
        tails_local_path: Optional[str] = None,
        tails_public_uri: Optional[str] = None,
        pending_pub: Sequence[str] = None,
        **kwargs,
    ):
        """Initialize the issuer revocation registry record."""
        super().__init__(
            record_id, state=state or IssuerRevRegRecord.STATE_INIT, **kwargs
        )
        self.cred_def_id = cred_def_id
        self.error_msg = error_msg
        self.issuer_did = issuer_did
        self.max_cred_num = max_cred_num or DEFAULT_REGISTRY_SIZE
        self.revoc_def_type = revoc_def_type or self.REVOC_DEF_TYPE_CL
        self.revoc_reg_id = revoc_reg_id
        self._revoc_reg_def = IndyRevRegDef.serde(revoc_reg_def)
        self._revoc_reg_entry = IndyRevRegEntry.serde(revoc_reg_entry)
        self.tag = tag
        self.tails_hash = tails_hash
        self.tails_local_path = tails_local_path
        self.tails_public_uri = tails_public_uri
        self.pending_pub = (
            sorted(set(pending_pub)) if pending_pub else []
        )  # order for eq comparison between instances

    @property
    def record_id(self) -> str:
        """Accessor for the record ID."""
        return self._id

    @property
    def revoc_reg_def(self) -> IndyRevRegDef:
        """Accessor; get deserialized."""
        return None if self._revoc_reg_def is None else self._revoc_reg_def.de

    @revoc_reg_def.setter
    def revoc_reg_def(self, value):
        """Setter; store de/serialized views."""
        self._revoc_reg_def = IndyRevRegDef.serde(value)

    @property
    def revoc_reg_entry(self) -> IndyRevRegEntry:
        """Accessor; get deserialized."""
        return None if self._revoc_reg_entry is None else self._revoc_reg_entry.de

    @revoc_reg_entry.setter
    def revoc_reg_entry(self, value):
        """Setter; store de/serialized views."""
        self._revoc_reg_entry = IndyRevRegEntry.serde(value)

    @property
    def record_value(self) -> Mapping:
        """Accessor for JSON value properties of this revocation registry record."""
        return {
            **{
                prop: getattr(self, prop)
                for prop in (
                    "error_msg",
                    "max_cred_num",
                    "tag",
                    "tails_hash",
                    "tails_public_uri",
                    "tails_local_path",
                    "pending_pub",
                )
            },
            **{
                prop: getattr(self, f"_{prop}").ser
                for prop in (
                    "revoc_reg_def",
                    "revoc_reg_entry",
                )
                if getattr(self, prop) is not None
            },
        }

    def _check_url(self, url) -> None:
        parsed = urlparse(url)
        if not (parsed.scheme and parsed.netloc and parsed.path):
            raise RevocationError("URI {} is not a valid URL".format(url))

    async def generate_registry(self, profile: Profile):
        """Create the revocation registry definition and tails file."""
        if not self.tag:
            self.tag = self._id or str(uuid4())

        if self.state != IssuerRevRegRecord.STATE_INIT:
            raise RevocationError(
                "Revocation registry {} in state {}: cannot generate".format(
                    self.revoc_reg_id, self.state
                )
            )

        issuer = profile.inject(IndyIssuer)
        tails_hopper_dir = indy_client_dir(join("tails", ".hopper"), create=True)

        LOGGER.debug("Creating revocation registry with size: %d", self.max_cred_num)

        try:
            (
                revoc_reg_id,
                revoc_reg_def_json,
                revoc_reg_entry_json,
            ) = await issuer.create_and_store_revocation_registry(
                self.issuer_did,
                self.cred_def_id,
                self.revoc_def_type,
                self.tag,
                self.max_cred_num,
                tails_hopper_dir,
            )
        except IndyIssuerError as err:
            raise RevocationError() from err

        if self.revoc_reg_id and revoc_reg_id != self.revoc_reg_id:
            raise RevocationError("Generated registry ID does not match assigned value")
        self.revoc_reg_id = revoc_reg_id
        self.revoc_reg_def = json.loads(revoc_reg_def_json)
        self.revoc_reg_entry = json.loads(revoc_reg_entry_json)
        self.state = IssuerRevRegRecord.STATE_GENERATED
        self.tails_hash = self.revoc_reg_def.value.tails_hash

        tails_dir = indy_client_dir(join("tails", self.revoc_reg_id), create=True)
        tails_path = join(tails_dir, self.tails_hash)
        move(join(tails_hopper_dir, self.tails_hash), tails_path)
        self.tails_local_path = tails_path

        async with profile.session() as session:
            await self.save(session, reason="Generated registry")

    async def set_tails_file_public_uri(self, profile: Profile, tails_file_uri: str):
        """Update tails file's publicly accessible URI."""
        if not (self.revoc_reg_def and self.revoc_reg_def.value.tails_location):
            raise RevocationError("Revocation registry undefined")

        self._check_url(tails_file_uri)

        self.tails_public_uri = tails_file_uri
        self._revoc_reg_def.de.value.tails_location = tails_file_uri  # update ...
        self.revoc_reg_def = self._revoc_reg_def.de  # ... and pick up change via setter
        async with profile.session() as session:
            await self.save(session, reason="Set tails file public URI")

    async def send_def(
        self,
        profile: Profile,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> dict:
        """Send the revocation registry definition to the ledger."""
        if not (self.revoc_reg_def and self.issuer_did):
            raise RevocationError(f"Revocation registry {self.revoc_reg_id} undefined")

        self._check_url(self.tails_public_uri)

        if self.state != IssuerRevRegRecord.STATE_GENERATED:
            raise RevocationError(
                "Revocation registry {} in state {}: cannot publish definition".format(
                    self.revoc_reg_id, self.state
                )
            )

        ledger = profile.inject(BaseLedger)
        async with ledger:
            rev_reg_res = await ledger.send_revoc_reg_def(
                self._revoc_reg_def.ser,
                self.issuer_did,
                write_ledger=write_ledger,
                endorser_did=endorser_did,
            )

        self.state = IssuerRevRegRecord.STATE_POSTED
        async with profile.session() as session:
            await self.save(session, reason="Published revocation registry definition")

        return rev_reg_res

    async def send_entry(
        self,
        profile: Profile,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> dict:
        """Send a registry entry to the ledger."""
        if not (
            self.revoc_reg_id
            and self.revoc_def_type
            and self.revoc_reg_entry
            and self.issuer_did
        ):
            raise RevocationError("Revocation registry undefined")

        self._check_url(self.tails_public_uri)

        if self.state not in (
            IssuerRevRegRecord.STATE_POSTED,
            IssuerRevRegRecord.STATE_ACTIVE,
            IssuerRevRegRecord.STATE_DECOMMISSIONED,
            IssuerRevRegRecord.STATE_FULL,  # can still publish revocation deltas
        ):
            raise RevocationError(
                "Revocation registry {} in state {}: cannot publish entry".format(
                    self.revoc_reg_id, self.state
                )
            )

        ledger = profile.inject(BaseLedger)
        async with ledger:
            try:
                rev_entry_res = await ledger.send_revoc_reg_entry(
                    self.revoc_reg_id,
                    self.revoc_def_type,
                    self._revoc_reg_entry.ser,
                    self.issuer_did,
                    write_ledger=write_ledger,
                    endorser_did=endorser_did,
                )
            except LedgerTransactionError as err:
                if "InvalidClientRequest" in err.roll_up:
                    # ... if the ledger write fails (with "InvalidClientRequest")
                    # e.g. acapy_agent.ledger.error.LedgerTransactionError:
                    #   Ledger rejected transaction request: client request invalid:
                    #   InvalidClientRequest(...)
                    # In this scenario we try to post a correction
                    LOGGER.warning("Retry ledger update/fix due to error")
                    LOGGER.warning(err)
                    (_, _, res) = await self.fix_ledger_entry(
                        profile,
                        True,
                        ledger.pool.genesis_txns,
                    )
                    rev_entry_res = {"result": res}
                    LOGGER.warning("Ledger update/fix applied")
                elif "InvalidClientTaaAcceptanceError" in err.roll_up:
                    # if no write access (with "InvalidClientTaaAcceptanceError")
                    # e.g. acapy_agent.ledger.error.LedgerTransactionError:
                    #   Ledger rejected transaction request: client request invalid:
                    #   InvalidClientTaaAcceptanceError(...)
                    LOGGER.error("Ledger update failed due to TAA issue")
                    LOGGER.error(err)
                    raise err
                else:
                    # not sure what happened, raise an error
                    LOGGER.error("Ledger update failed due to unknown issue")
                    LOGGER.error(err)
                    raise err
        if self.state == IssuerRevRegRecord.STATE_POSTED:
            self.state = IssuerRevRegRecord.STATE_ACTIVE  # initial entry activates
            async with profile.session() as session:
                await self.save(
                    session, reason="Published initial revocation registry entry"
                )

        return rev_entry_res

    def _get_revoked_discrepancies(
        self, recs: Sequence[IssuerCredRevRecord], rev_reg_delta: dict
    ) -> Tuple[list, int]:
        revoked_ids = []
        rec_count = 0
        for rec in recs:
            if rec.state == IssuerCredRevRecord.STATE_REVOKED:
                revoked_ids.append(int(rec.cred_rev_id))
                if int(rec.cred_rev_id) not in rev_reg_delta["value"]["revoked"]:
                    rec_count += 1

        return revoked_ids, rec_count

    async def fix_ledger_entry(
        self,
        profile: Profile,
        apply_ledger_update: bool,
        genesis_transactions: str,
    ) -> Tuple[dict, dict, dict]:
        """Fix the ledger entry to match wallet-recorded credentials."""
        recovery_txn = {}
        applied_txn = {}

        # get rev reg delta (revocations published to ledger)
        ledger = profile.inject(BaseLedger)
        async with ledger:
            (rev_reg_delta, _) = await ledger.get_revoc_reg_delta(self.revoc_reg_id)

        # get rev reg records from wallet (revocations and status)
        async with profile.session() as session:
            recs = await IssuerCredRevRecord.query_by_ids(
                session, rev_reg_id=self.revoc_reg_id
            )

        revoked_ids, rec_count = self._get_revoked_discrepancies(recs, rev_reg_delta)

        LOGGER.debug(f"Fixed entry recs count = {rec_count}")
        LOGGER.debug(f"Rev reg entry value: {self.revoc_reg_entry.value}")
        LOGGER.debug(f"Rev reg delta: {rev_reg_delta.get('value')}")

        # No update required if no discrepancies
        if rec_count == 0:
            return (rev_reg_delta, {}, {})

        # We have revocation discrepancies, generate the recovery txn
        async with profile.session() as session:
            # We need the cred_def and rev_reg_def_private to generate the recovery txn
            issuer_rev_reg_record = await IssuerRevRegRecord.retrieve_by_revoc_reg_id(
                session, self.revoc_reg_id
            )
            cred_def_id = issuer_rev_reg_record.cred_def_id
            cred_def = await session.handle.fetch(CATEGORY_CRED_DEF, cred_def_id)
            rev_reg_def_private = await session.handle.fetch(
                CATEGORY_REV_REG_DEF_PRIVATE, self.revoc_reg_id
            )

        credx_module = importlib.import_module("indy_credx")
        cred_defn = credx_module.CredentialDefinition.load(cred_def.value_json)
        rev_reg_defn_private = credx_module.RevocationRegistryDefinitionPrivate.load(
            rev_reg_def_private.value_json
        )
        calculated_txn = await generate_ledger_rrrecovery_txn(
            genesis_transactions,
            self.revoc_reg_id,
            revoked_ids,
            cred_defn,
            rev_reg_defn_private,
        )
        recovery_txn = json.loads(calculated_txn.to_json())

        LOGGER.debug(f"Applying ledger update: {apply_ledger_update}")
        if apply_ledger_update:
            async with profile.session() as session:
                ledger = session.inject_or(BaseLedger)
                if not ledger:
                    reason = "No ledger available"
                    if not session.context.settings.get_value("wallet.type"):
                        reason += ": missing wallet-type?"
                    raise LedgerError(reason=reason)

                async with ledger:
                    ledger_response = await ledger.send_revoc_reg_entry(
                        self.revoc_reg_id, "CL_ACCUM", recovery_txn
                    )

            applied_txn = ledger_response["result"]

            # Update the local wallets rev reg entry with the new accumulator value
            async with profile.session() as session:
                rev_reg = await session.handle.fetch(
                    CATEGORY_REV_REG, self.revoc_reg_id, for_update=True
                )
                new_value_json = rev_reg.value_json
                new_value_json["value"]["accum"] = applied_txn["txn"]["data"]["value"][
                    "accum"
                ]
                await session.handle.replace(
                    CATEGORY_REV_REG,
                    rev_reg.name,
                    json.dumps(new_value_json),
                    rev_reg.tags,
                )

        return (rev_reg_delta, recovery_txn, applied_txn)

    @property
    def has_local_tails_file(self) -> bool:
        """Check if a local copy of the tails file is available."""
        return bool(self.tails_local_path) and Path(self.tails_local_path).is_file()

    async def upload_tails_file(self, profile: Profile):
        """Upload the local tails file to the tails server."""
        tails_server = IndyTailsServer()
        if not self.has_local_tails_file:
            raise RevocationError("Local tails file not found")

        (upload_success, result) = await tails_server.upload_tails_file(
            profile.context,
            self.revoc_reg_id,
            self.tails_local_path,
            interval=0.8,
            backoff=-0.5,
            max_attempts=5,  # heuristic: respect HTTP timeout
        )
        if not upload_success:
            raise RevocationError(
                f"Tails file for rev reg {self.revoc_reg_id} failed to upload: {result}"
            )
        await self.set_tails_file_public_uri(profile, result)

    async def mark_pending(self, session: ProfileSession, cred_rev_id: str) -> None:
        """Mark a credential revocation id as revoked pending publication to ledger.

        Args:
            session: The profile session to use
            cred_rev_id: The credential revocation identifier for credential to revoke
        """
        if cred_rev_id not in self.pending_pub:
            self.pending_pub.append(cred_rev_id)
            self.pending_pub.sort()

        await self.save(session, reason="Marked pending revocation")

    async def clear_pending(
        self, session: ProfileSession, cred_rev_ids: Sequence[str] = None
    ) -> None:
        """Clear pending revocations and save any resulting record change.

        Args:
            session: The profile session to use
            cred_rev_ids: Credential revocation identifiers to clear; default all
        """
        if self.pending_pub:
            if cred_rev_ids:
                self.pending_pub = [r for r in self.pending_pub if r not in cred_rev_ids]
            else:
                self.pending_pub.clear()
            await self.save(session, reason="Cleared pending revocations")

    def get_registry(self) -> RevocationRegistry:
        """Create a `RevocationRegistry` instance from this record."""
        return RevocationRegistry(
            self.revoc_reg_id,
            cred_def_id=self.cred_def_id,
            issuer_did=self.issuer_did,
            max_creds=self.max_cred_num,
            reg_def_type=self.revoc_def_type,
            tag=self.tag,
            tails_local_path=self.tails_local_path,
            tails_public_uri=self.tails_public_uri,
            tails_hash=self.tails_hash,
        )

    @classmethod
    async def query_by_cred_def_id(
        cls,
        session: ProfileSession,
        cred_def_id: str,
        state: Optional[str] = None,
        negative_state: Optional[str] = None,
        limit=None,
    ) -> Sequence["IssuerRevRegRecord"]:
        """Retrieve issuer revocation registry records by credential definition ID.

        Args:
            session: The profile session to use
            cred_def_id: The credential definition ID to filter by
            state: A state value to filter by
            negative_state: A state value to exclude
            limit: The maximum number of records to return
        """
        tag_filter = dict(
            filter(
                lambda f: f[1] is not None,
                (("cred_def_id", cred_def_id), ("state", state)),
            )
        )
        return await cls.query(
            session,
            tag_filter,
            post_filter_positive={"state": state} if state else None,
            post_filter_negative={"state": negative_state} if negative_state else None,
            limit=limit,
        )

    @classmethod
    async def query_by_pending(
        cls, session: ProfileSession
    ) -> Sequence["IssuerRevRegRecord"]:
        """Retrieve issuer revocation records with revocations pending.

        Args:
            session: The profile session to use
        """
        return await cls.query(
            session=session,
            tag_filter=None,
            post_filter_positive=None,
            post_filter_negative={"pending_pub": []},
        )

    @classmethod
    async def retrieve_by_revoc_reg_id(
        cls, session: ProfileSession, revoc_reg_id: str, for_update: bool = False
    ) -> "IssuerRevRegRecord":
        """Retrieve a revocation registry record by revocation registry ID.

        Args:
            session: The profile session to use
            revoc_reg_id: The revocation registry ID
            for_update: Retrieve for update
        """
        tag_filter = {"revoc_reg_id": revoc_reg_id}
        return await cls.retrieve_by_tag_filter(
            session, tag_filter, for_update=for_update
        )

    async def set_state(self, session: ProfileSession, state: Optional[str] = None):
        """Change the registry state (default full)."""
        self.state = state or IssuerRevRegRecord.STATE_FULL
        await self.save(
            session, reason=f"Marked rev reg {self.revoc_reg_id} as {self.state}"
        )

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)

    def __lt__(self, other):
        """Order by creation time."""
        return (self.created_at or 0) < (other.created_at or 0)


class IssuerRevRegRecordSchema(BaseRecordSchema):
    """Schema to allow serialization/deserialization of issuer rev reg records."""

    class Meta:
        """IssuerRevRegRecordSchema metadata."""

        model_class = IssuerRevRegRecord

    record_id = fields.Str(
        required=False,
        metadata={
            "description": "Issuer revocation registry record identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    state = fields.Str(
        required=False,
        metadata={
            "description": "Issue revocation registry record state",
            "example": IssuerRevRegRecord.STATE_ACTIVE,
        },
    )
    cred_def_id = fields.Str(
        required=False,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    error_msg = fields.Str(
        required=False,
        metadata={
            "description": "Error message",
            "example": "Revocation registry undefined",
        },
    )
    issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Issuer DID", "example": INDY_DID_EXAMPLE},
    )
    max_cred_num = fields.Int(
        required=False,
        metadata={
            "description": "Maximum number of credentials for revocation registry",
            "strict": True,
            "example": 1000,
        },
    )
    revoc_def_type = fields.Str(
        required=False,
        validate=validate.Equal("CL_ACCUM"),
        metadata={
            "description": "Revocation registry type (specify CL_ACCUM)",
            "example": "CL_ACCUM",
        },
    )
    revoc_reg_id = fields.Str(
        required=False,
        validate=INDY_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Revocation registry identifier",
            "example": INDY_REV_REG_ID_EXAMPLE,
        },
    )
    revoc_reg_def = fields.Nested(
        IndyRevRegDefSchema(),
        required=False,
        metadata={"description": "Revocation registry definition"},
    )
    revoc_reg_entry = fields.Nested(
        IndyRevRegEntrySchema(),
        required=False,
        metadata={"description": "Revocation registry entry"},
    )
    tag = fields.Str(
        required=False,
        metadata={"description": "Tag within issuer revocation registry identifier"},
    )
    tails_hash = fields.Str(
        required=False,
        validate=BASE58_SHA256_HASH_VALIDATE,
        metadata={"description": "Tails hash", "example": BASE58_SHA256_HASH_EXAMPLE},
    )
    tails_public_uri = fields.Str(
        required=False, metadata={"description": "Public URI for tails file"}
    )
    tails_local_path = fields.Str(
        required=False, metadata={"description": "Local path to tails file"}
    )
    pending_pub = fields.List(
        fields.Str(metadata={"example": "23"}),
        required=False,
        metadata={
            "description": (
                "Credential revocation identifier for credential revoked and pending"
                " publication to ledger"
            )
        },
    )



"""Problem Details objects for error handling."""

# https://www.w3.org/TR/vc-data-integrity/#processing-errors
PROBLEM_DETAILS = {
    "PROOF_GENERATION_ERROR": {
        "type": "https://w3id.org/security#PROOF_GENERATION_ERROR"
    },
    "PROOF_VERIFICATION_ERROR": {
        "type": "https://w3id.org/security#PROOF_VERIFICATION_ERROR"
    },
    "PROOF_TRANSFORMATION_ERROR": {
        "type": "https://w3id.org/security#PROOF_TRANSFORMATION_ERROR"
    },
    "INVALID_DOMAIN_ERROR": {"type": "https://w3id.org/security#INVALID_DOMAIN_ERROR"},
    "INVALID_CHALLENGE_ERROR": {
        "type": "https://w3id.org/security#INVALID_CHALLENGE_ERROR"
    },
}



"""Admin server classes."""

import asyncio
import logging
import re
import warnings
import weakref
from typing import Callable, Coroutine, Optional

import aiohttp_cors
import jwt
from aiohttp import web
from aiohttp_apispec import setup_aiohttp_apispec, validation_middleware
from uuid_utils import uuid4

from ..config.injection_context import InjectionContext
from ..config.logging import context_wallet_id
from ..core.event_bus import Event, EventBus
from ..core.plugin_registry import PluginRegistry
from ..core.profile import Profile
from ..ledger.error import LedgerConfigError, LedgerTransactionError
from ..messaging.responder import BaseResponder
from ..multitenant.base import BaseMultitenantManager
from ..multitenant.error import InvalidTokenError, MultitenantManagerError
from ..storage.base import BaseStorage
from ..storage.error import StorageNotFoundError
from ..storage.type import RECORD_TYPE_ACAPY_UPGRADING
from ..transport.outbound.message import OutboundMessage
from ..transport.outbound.status import OutboundSendStatus
from ..transport.queue.basic import BasicMessageQueue
from ..utils import general as general_utils
from ..utils.extract_validation_error import extract_validation_error_message
from ..utils.server import remove_unwanted_headers
from ..utils.stats import Collector
from ..utils.task_queue import TaskQueue
from ..version import __version__
from ..wallet import singletons
from ..wallet.anoncreds_upgrade import check_upgrade_completion_loop
from .base_server import BaseAdminServer
from .error import AdminSetupError
from .request_context import AdminRequestContext
from .routes import (
    config_handler,
    liveliness_handler,
    plugins_handler,
    readiness_handler,
    redirect_handler,
    shutdown_handler,
    status_handler,
    status_reset_handler,
)

LOGGER = logging.getLogger(__name__)

EVENT_PATTERN_WEBHOOK = re.compile("^acapy::webhook::(.*)$")
EVENT_PATTERN_RECORD = re.compile("^acapy::record::([^:]*)(?:::.*)?$")

EVENT_WEBHOOK_MAPPING = {
    "acapy::basicmessage::received": "basicmessages",
    "acapy::problem_report": "problem_report",
    "acapy::ping::received": "ping",
    "acapy::ping::response_received": "ping",
    "acapy::actionmenu::received": "actionmenu",
    "acapy::actionmenu::get-active-menu": "get-active-menu",
    "acapy::actionmenu::perform-menu-action": "perform-menu-action",
    "acapy::keylist::updated": "keylist",
}

anoncreds_wallets = singletons.IsAnonCredsSingleton().wallets
in_progress_upgrades = singletons.UpgradeInProgressSingleton()

status_paths = ("/status/live", "/status/ready")


class AdminResponder(BaseResponder):
    """Handle outgoing messages from message handlers."""

    def __init__(
        self,
        profile: Profile,
        send: Coroutine,
        **kwargs,
    ):
        """Initialize an instance of `AdminResponder`.

        Args:
            profile (Profile): The profile for this responder.
            send (Coroutine): Function to send outbound message.
            **kwargs: Additional keyword arguments.

        """
        super().__init__(**kwargs)
        # Weakly hold the profile so this reference doesn't prevent profiles
        # from being cleaned up when appropriate.
        # Binding this AdminResponder to the profile's context creates a circular
        # reference.
        self._profile = weakref.ref(profile)
        self._send = send

    async def send_outbound(
        self, message: OutboundMessage, **kwargs
    ) -> OutboundSendStatus:
        """Send outbound message.

        Args:
            message: The `OutboundMessage` to be sent
            **kwargs: Additional keyword arguments
        """
        profile = self._profile()
        if not profile:
            raise RuntimeError("weakref to profile has expired")
        return await self._send(profile, message)

    async def send_webhook(self, topic: str, payload: dict):
        """Dispatch a webhook. DEPRECATED: use the event bus instead.

        Args:
            topic: the webhook topic identifier
            payload: the webhook payload value
        """
        warnings.warn(
            "responder.send_webhook is deprecated; please use the event bus instead.",
            DeprecationWarning,
        )
        profile = self._profile()
        if not profile:
            raise RuntimeError("weakref to profile has expired")
        await profile.notify("acapy::webhook::" + topic, payload)

    @property
    def send_fn(self) -> Coroutine:
        """Accessor for async function to send outbound message."""
        return self._send


@web.middleware
async def ready_middleware(request: web.BaseRequest, handler: Coroutine):
    """Only continue if application is ready to take work."""

    is_status_check = str(request.rel_url).rstrip("/") in status_paths
    is_app_ready = request.app._state.get("ready")

    if not (is_status_check or is_app_ready):
        raise web.HTTPServiceUnavailable(reason="Shutdown in progress")

    try:
        return await handler(request)
    except web.HTTPFound as e:
        # redirect, typically / -> /api/doc
        LOGGER.info("Handler redirect to: %s", e.location)
        raise
    except asyncio.CancelledError:
        # redirection spawns new task and cancels old
        LOGGER.debug("Task cancelled")
        raise
    except (web.HTTPUnauthorized, jwt.InvalidTokenError, InvalidTokenError) as e:
        LOGGER.info(
            "Unauthorized access during %s %s: %s", request.method, request.path, e
        )
        raise web.HTTPUnauthorized(reason=str(e)) from e
    except (web.HTTPBadRequest, MultitenantManagerError) as e:
        LOGGER.info("Bad request during %s %s: %s", request.method, request.path, e)
        raise web.HTTPBadRequest(reason=str(e)) from e
    except (web.HTTPNotFound, StorageNotFoundError) as e:
        LOGGER.info(
            "Not Found error occurred during %s %s: %s",
            request.method,
            request.path,
            e,
        )
        raise web.HTTPNotFound(reason=str(e)) from e
    except web.HTTPUnprocessableEntity as e:
        validation_error_message = extract_validation_error_message(e)
        LOGGER.info(
            "Unprocessable Entity occurred during %s %s: %s",
            request.method,
            request.path,
            validation_error_message,
        )
        raise web.HTTPUnprocessableEntity(reason=validation_error_message) from e
    except (LedgerConfigError, LedgerTransactionError) as e:
        # fatal, signal server shutdown
        LOGGER.critical("Shutdown with %s", str(e))
        request.app._state["ready"] = False
        request.app._state["alive"] = False
        raise
    except Exception as e:
        LOGGER.exception("Handler error with exception:", exc_info=e)
        raise


@web.middleware
async def upgrade_middleware(request: web.BaseRequest, handler: Coroutine):
    """Blocking middleware for upgrades."""
    # Skip upgrade check for status checks
    if str(request.rel_url).startswith("/status/"):
        return await handler(request)

    context: AdminRequestContext = request["context"]

    # Already upgraded
    if context.profile.name in anoncreds_wallets:
        return await handler(request)

    # Upgrade in progress
    if context.profile.name in in_progress_upgrades.wallets:
        raise web.HTTPServiceUnavailable(reason="Upgrade in progress")

    # Avoid try/except in middleware with find_all_records
    upgrade_initiated = []
    async with context.profile.session() as session:
        storage = session.inject(BaseStorage)
        upgrade_initiated = await storage.find_all_records(RECORD_TYPE_ACAPY_UPGRADING)
        if upgrade_initiated:
            # If we get here, than another instance started an upgrade
            # We need to check for completion (or fail) in another process
            in_progress_upgrades.set_wallet(context.profile.name)
            is_subwallet = context.metadata and "wallet_id" in context.metadata
            asyncio.create_task(
                check_upgrade_completion_loop(
                    context.profile,
                    is_subwallet,
                )
            )
            raise web.HTTPServiceUnavailable(reason="Upgrade in progress")

    return await handler(request)


@web.middleware
async def debug_middleware(request: web.BaseRequest, handler: Coroutine):
    """Show request detail in debug log."""

    if LOGGER.isEnabledFor(logging.DEBUG):  # Skipped if DEBUG is not enabled
        LOGGER.debug("Incoming request: %s %s", request.method, request.path_qs)
        is_status_check = str(request.rel_url).startswith("/status/")
        if not is_status_check:  # Don't log match info for status checks; reduces noise
            LOGGER.debug("Match info: %s", request.match_info)

            if request.body_exists:  # Only log body if it exists
                LOGGER.debug("Body: %s", await request.text())

    return await handler(request)


class AdminServer(BaseAdminServer):
    """Admin HTTP server class."""

    def __init__(
        self,
        host: str,
        port: int,
        context: InjectionContext,
        root_profile: Profile,
        outbound_message_router: Coroutine,
        webhook_router: Callable,
        conductor_stop: Coroutine,
        task_queue: Optional[TaskQueue] = None,
        conductor_stats: Optional[Coroutine] = None,
    ):
        """Initialize an AdminServer instance.

        Args:
            host (str): The host to listen on.
            port (int): The port to listen on.
            context (InjectionContext): The application context instance.
            root_profile (Profile): The root profile.
            outbound_message_router (Coroutine): Coroutine for delivering
                outbound messages.
            webhook_router (Callable): Callable for delivering webhooks.
            conductor_stop (Coroutine): Conductor (graceful) stop for shutdown API call.
            task_queue (TaskQueue, optional): An optional task queue for handlers.
            conductor_stats (Coroutine, optional): Conductor statistics API call.
        """
        self.app = None
        self.admin_api_key = context.settings.get("admin.admin_api_key")
        self.admin_insecure_mode = bool(context.settings.get("admin.admin_insecure_mode"))
        self.host = host
        self.port = port
        self.context = context
        self.conductor_stop = conductor_stop
        self.conductor_stats = conductor_stats
        self.loaded_modules = []
        self.outbound_message_router = outbound_message_router
        self.root_profile = root_profile
        self.task_queue = task_queue
        self.webhook_router = webhook_router
        self.websocket_queues = {}
        self.site = None
        self.multitenant_manager = context.inject_or(BaseMultitenantManager)

    async def make_application(self) -> web.Application:
        """Get the aiohttp application instance."""

        middlewares = [ready_middleware, debug_middleware]

        # admin-token and admin-token are mutually exclusive and required.
        # This should be enforced during parameter parsing but to be sure,
        # we check here.
        assert self.admin_insecure_mode ^ bool(self.admin_api_key)

        collector = self.context.inject_or(Collector)

        @web.middleware
        async def setup_context(request: web.Request, handler):
            authorization_header = request.headers.get("Authorization")
            profile = self.root_profile
            meta_data = {}
            # Multitenancy context setup
            if self.multitenant_manager and authorization_header:
                try:
                    bearer, _, token = authorization_header.partition(" ")
                    if bearer != "Bearer":
                        raise web.HTTPUnauthorized(
                            reason="Invalid Authorization header structure"
                        )

                    profile = await self.multitenant_manager.get_profile_for_token(
                        self.context, token
                    )
                    (
                        walletid,
                        walletkey,
                    ) = self.multitenant_manager.get_wallet_details_from_token(
                        token=token
                    )
                    wallet_id = profile.settings.get("wallet.id")
                    context_wallet_id.set(wallet_id)
                    meta_data = {
                        "wallet_id": walletid,
                        "wallet_key": walletkey,
                    }
                except MultitenantManagerError as err:
                    raise web.HTTPUnauthorized(reason=err.roll_up)
                except (jwt.InvalidTokenError, StorageNotFoundError):
                    raise web.HTTPUnauthorized()

            # Create a responder with the request specific context
            responder = AdminResponder(
                profile,
                self.outbound_message_router,
            )
            profile.context.injector.bind_instance(BaseResponder, responder)

            # TODO may dynamically adjust the profile used here according to
            # headers or other parameters
            if self.multitenant_manager and authorization_header:
                admin_context = AdminRequestContext(
                    profile=profile,
                    root_profile=self.root_profile,
                    metadata=meta_data,
                )
            else:
                admin_context = AdminRequestContext(
                    profile=profile,
                )

            request["context"] = admin_context
            request["outbound_message_router"] = responder.send

            if collector:
                handler = collector.wrap_coro(handler, [handler.__qualname__])
            if self.task_queue:
                task = await self.task_queue.put(handler(request))
                return await task
            return await handler(request)

        middlewares.append(setup_context)

        # Upgrade middleware needs the context setup
        middlewares.append(upgrade_middleware)

        # Register validation_middleware last avoiding unauthorized validations
        middlewares.append(validation_middleware)

        app = web.Application(
            middlewares=middlewares,
            client_max_size=(
                self.context.settings.get("admin.admin_client_max_request_size", 1)
                * 1024
                * 1024
            ),
        )

        server_routes = [
            web.get("/", redirect_handler, allow_head=True),
            web.get("/plugins", plugins_handler, allow_head=False),
            web.get("/status", status_handler, allow_head=False),
            web.get("/status/config", config_handler, allow_head=False),
            web.post("/status/reset", status_reset_handler),
            web.get("/status/live", liveliness_handler, allow_head=False),
            web.get("/status/ready", readiness_handler, allow_head=False),
            web.get("/shutdown", shutdown_handler, allow_head=False),
            web.get("/ws", self.websocket_handler, allow_head=False),
        ]
        app.add_routes(server_routes)

        app.on_response_prepare.append(remove_unwanted_headers)

        plugin_registry = self.context.inject_or(PluginRegistry)
        if plugin_registry:
            await plugin_registry.register_admin_routes(app)

        cors = aiohttp_cors.setup(
            app,
            defaults={
                "*": aiohttp_cors.ResourceOptions(
                    allow_credentials=True,
                    expose_headers="*",
                    allow_headers="*",
                    allow_methods="*",
                )
            },
        )
        for route in app.router.routes():
            cors.add(route)
        # get agent label
        agent_label = self.context.settings.get("default_label")
        version_string = f"v{__version__}"

        setup_aiohttp_apispec(
            app=app, title=agent_label, version=version_string, swagger_path="/api/doc"
        )
        app.on_startup.append(self.on_startup)

        # ensure we always have status values
        app._state["ready"] = False
        app._state["alive"] = False

        # set global-like variables
        app["context"] = self.context
        app["conductor_stats"] = self.conductor_stats
        app["conductor_stop"] = self.conductor_stop

        return app

    async def start(self) -> None:
        """Start the webserver.

        Raises:
            AdminSetupError: If there was an error starting the webserver

        """

        def sort_dict(raw: dict) -> dict:
            """Order (JSON, string keys) dict asciibetically by key, recursively."""
            for k, v in raw.items():
                if isinstance(v, dict):
                    raw[k] = sort_dict(v)
            return dict(sorted(raw.items(), key=lambda x: x[0]))

        self.app = await self.make_application()
        runner = web.AppRunner(self.app)
        await runner.setup()

        plugin_registry = self.context.inject_or(PluginRegistry)
        if plugin_registry:
            plugin_registry.post_process_routes(self.app)

        event_bus = self.context.inject_or(EventBus)
        if event_bus:
            event_bus.subscribe(EVENT_PATTERN_WEBHOOK, self._on_webhook_event)
            event_bus.subscribe(EVENT_PATTERN_RECORD, self._on_record_event)

            # Only include forward webhook events if the option is enabled
            if self.context.settings.get_bool("monitor_forward", False):
                EVENT_WEBHOOK_MAPPING["acapy::forward::received"] = "forward"

            for event_topic, webhook_topic in EVENT_WEBHOOK_MAPPING.items():
                event_bus.subscribe(
                    re.compile(re.escape(event_topic)),
                    lambda profile, event, webhook_topic=webhook_topic: self.send_webhook(
                        profile, webhook_topic, event.payload
                    ),
                )

        # order tags alphabetically, parameters deterministically and pythonically
        swagger_dict = self.app._state["swagger_dict"]
        swagger_dict.get("tags", []).sort(key=lambda t: t["name"])

        # sort content per path and sort paths
        for path_spec in swagger_dict["paths"].values():
            for method_spec in path_spec.values():
                method_spec["parameters"].sort(
                    key=lambda p: (p["in"], not p["required"], p["name"])
                )
        for path in sorted(swagger_dict["paths"]):
            swagger_dict["paths"][path] = swagger_dict["paths"].pop(path)

        # order definitions alphabetically by dict key
        swagger_dict["definitions"] = sort_dict(swagger_dict["definitions"])

        self.site = web.TCPSite(runner, host=self.host, port=self.port)

        try:
            await self.site.start()
            self.app._state["ready"] = True
            self.app._state["alive"] = True
        except OSError:
            raise AdminSetupError(
                "Unable to start webserver with host "
                + f"'{self.host}' and port '{self.port}'\n"
            )

    async def stop(self) -> None:
        """Stop the webserver."""
        # Stopped before admin server is created
        if not self.app:
            return

        self.app._state["ready"] = False  # in case call does not come through OpenAPI
        for queue in self.websocket_queues.values():
            queue.stop()
        if self.site:
            await self.site.stop()
            self.site = None

    async def on_startup(self, app: web.Application):
        """Perform webserver startup actions."""
        security_definitions = {}
        security = []

        if self.admin_api_key:
            security_definitions["ApiKeyHeader"] = {
                "type": "apiKey",
                "in": "header",
                "name": "X-API-KEY",
            }
            security.append({"ApiKeyHeader": []})
        if self.multitenant_manager:
            security_definitions["AuthorizationHeader"] = {
                "type": "apiKey",
                "in": "header",
                "name": "Authorization",
                "description": "Bearer token. Be sure to prepend token with 'Bearer '",
            }

            # If multitenancy is enabled we need Authorization header
            multitenant_security = {"AuthorizationHeader": []}
            # If admin api key is also enabled, we need both for subwallet requests
            if self.admin_api_key:
                multitenant_security["ApiKeyHeader"] = []
            security.append(multitenant_security)

        if self.admin_api_key or self.multitenant_manager:
            swagger = app["swagger_dict"]
            swagger["securityDefinitions"] = security_definitions
            swagger["security"] = security

    def notify_fatal_error(self):
        """Set our readiness flags to force a restart (openshift)."""
        LOGGER.error("Received shutdown request notify_fatal_error()")
        self.app._state["ready"] = False
        self.app._state["alive"] = False

    async def websocket_handler(self, request):
        """Send notifications to admin client over websocket."""

        ws = web.WebSocketResponse()
        await ws.prepare(request)
        socket_id = str(uuid4())
        queue = BasicMessageQueue()
        loop = asyncio.get_event_loop()

        if self.admin_insecure_mode:
            # open to send websocket messages without api key auth
            queue.authenticated = True
        else:
            header_admin_api_key = request.headers.get("x-api-key")
            # authenticated via http header?
            queue.authenticated = general_utils.const_compare(
                header_admin_api_key, self.admin_api_key
            )

        try:
            self.websocket_queues[socket_id] = queue
            await queue.enqueue(
                {
                    "topic": "settings",
                    "payload": {
                        "authenticated": queue.authenticated,
                        "label": self.context.settings.get("default_label"),
                        "endpoint": self.context.settings.get("default_endpoint"),
                        "no_receive_invites": self.context.settings.get(
                            "admin.no_receive_invites", False
                        ),
                        "help_link": self.context.settings.get("admin.help_link"),
                    },
                }
            )

            closed = False
            receive = loop.create_task(ws.receive_json())
            send = loop.create_task(queue.dequeue(timeout=5.0))

            while not closed:
                try:
                    await asyncio.wait(
                        (receive, send), return_when=asyncio.FIRST_COMPLETED
                    )
                    if ws.closed:
                        closed = True

                    if receive.done():
                        if not closed:
                            msg_received = None
                            msg_api_key = None
                            try:
                                # this call can re-raise exceptions from inside the task
                                msg_received = receive.result()
                                msg_api_key = msg_received.get("x-api-key")
                            except Exception:
                                LOGGER.exception("Exception in websocket receiving task:")
                            if self.admin_api_key and general_utils.const_compare(
                                self.admin_api_key, msg_api_key
                            ):
                                # authenticated via websocket message
                                queue.authenticated = True

                            receive = loop.create_task(ws.receive_json())

                    if send.done():
                        try:
                            msg = send.result()
                        except asyncio.TimeoutError:
                            msg = None

                        if msg is None:
                            # we send fake pings because the JS client
                            # can't detect real ones
                            msg = {
                                "topic": "ping",
                                "authenticated": queue.authenticated,
                            }
                        if not closed:
                            if msg:
                                await ws.send_json(msg)
                            send = loop.create_task(queue.dequeue(timeout=5.0))

                except asyncio.CancelledError:
                    closed = True

            if not receive.done():
                receive.cancel()
            if not send.done():
                send.cancel()

        finally:
            del self.websocket_queues[socket_id]

        return ws

    async def _on_webhook_event(self, profile: Profile, event: Event):
        match = EVENT_PATTERN_WEBHOOK.search(event.topic)
        webhook_topic = match.group(1) if match else None
        if webhook_topic:
            await self.send_webhook(profile, webhook_topic, event.payload)

    async def _on_record_event(self, profile: Profile, event: Event):
        match = EVENT_PATTERN_RECORD.search(event.topic)
        webhook_topic = match.group(1) if match else None
        if webhook_topic:
            await self.send_webhook(profile, webhook_topic, event.payload)

    async def send_webhook(
        self, profile: Profile, topic: str, payload: Optional[dict] = None
    ):
        """Add a webhook to the queue, to send to all registered targets."""
        wallet_id = profile.settings.get("wallet.id")
        webhook_urls = profile.settings.get("admin.webhook_urls")

        metadata = None
        if wallet_id:
            metadata = {"x-wallet-id": wallet_id}

        if self.webhook_router:
            for endpoint in webhook_urls:
                self.webhook_router(
                    topic,
                    payload,
                    endpoint,
                    None,
                    metadata,
                )

        # set ws webhook body, optionally add wallet id for multitenant mode
        webhook_body = {"topic": topic, "payload": payload}
        if wallet_id:
            webhook_body["wallet_id"] = wallet_id

        for queue in self.websocket_queues.values():
            if queue.authenticated or topic in ("ping", "settings"):
                await queue.enqueue(webhook_body)



from unittest import IsolatedAsyncioTestCase

from ..base import InjectionError
from ..injection_context import InjectionContext


class TestInjectionContext(IsolatedAsyncioTestCase):
    def setUp(self):
        self.test_key = "TEST"
        self.test_value = "VALUE"
        self.test_scope = "SCOPE"
        self.test_settings = {self.test_key: self.test_value}
        self.test_instance = InjectionContext(settings=self.test_settings)

    def test_settings_init(self):
        """Test settings initialization."""
        for key in self.test_settings:
            assert key in self.test_instance.settings
            assert self.test_instance.settings[key] == self.test_settings[key]

    def test_settings_scope(self):
        """Test scoped settings."""
        upd_settings = {self.test_key: "NEWVAL"}
        context = self.test_instance.start_scope(upd_settings)
        assert context.settings[self.test_key] == "NEWVAL"
        assert self.test_instance.settings[self.test_key] == self.test_value

        context.settings = upd_settings
        assert context.settings == upd_settings

    async def test_inject_simple(self):
        """Test a basic injection."""
        assert self.test_instance.inject_or(str) is None
        with self.assertRaises(InjectionError):
            self.test_instance.inject(str)
        self.test_instance.injector.bind_instance(str, self.test_value)
        assert self.test_instance.inject(str) is self.test_value

        self.test_instance.injector = None
        assert self.test_instance.injector is None

    async def test_inject_scope(self):
        """Test a scoped injection."""
        context = self.test_instance.start_scope()
        assert context.inject_or(str) is None
        context.injector.bind_instance(str, self.test_value)
        assert context.inject(str) is self.test_value
        assert self.test_instance.inject_or(str) is None



from unittest import TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import PROTOCOL_PACKAGE, QUERY
from ..query import Query


class TestQuery(TestCase):
    test_query = "*"
    test_comment = "comment"

    def test_init(self):
        query = Query(query=self.test_query, comment=self.test_comment)
        assert query.query == self.test_query
        assert query.comment == self.test_comment

    def test_type(self):
        query = Query(query=self.test_query, comment=self.test_comment)
        assert query._type == DIDCommPrefix.qualify_current(QUERY)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.query.QuerySchema.load")
    def test_deserialize(self, mock_query_schema_load):
        obj = {"obj": "obj"}

        query = Query.deserialize(obj)
        mock_query_schema_load.assert_called_once_with(obj)

        assert query is mock_query_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.query.QuerySchema.dump")
    def test_serialize(self, mock_query_schema_dump):
        query = Query(query=self.test_query, comment=self.test_comment)

        query_dict = query.serialize()
        mock_query_schema_dump.assert_called_once_with(query)

        assert query_dict is mock_query_schema_dump.return_value


class TestQuerySchema(TestCase):
    query = Query(query="*", comment="comment")

    def test_make_model(self):
        data = self.query.serialize()
        model_instance = Query.deserialize(data)
        assert isinstance(model_instance, Query)



from unittest import IsolatedAsyncioTestCase

import pytest
from aries_askar import AskarError, AskarErrorCode

from acapy_agent.indy.credx.holder import CredxError, IndyHolderError
from acapy_agent.utils.testing import create_test_profile

from ....tests import mock
from .. import holder


@pytest.mark.askar
@pytest.mark.indy_credx
class TestIndyCredxGetLinkSecret(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.holder_profile = await create_test_profile()
        self.holder = holder.IndyCredxHolder(self.holder_profile)

        self.mock_session = mock.MagicMock()
        self.mock_session.__aenter__.return_value = self.mock_session
        self.mock_session.__aexit__.return_value = None
        self.holder._profile.session = mock.MagicMock(return_value=self.mock_session)

    @mock.patch("acapy_agent.indy.credx.holder.LinkSecret.load")
    @mock.patch("acapy_agent.indy.credx.holder.LinkSecret.create")
    async def test_get_link_secret_existing(self, mock_create, mock_load):
        # Mock session and record
        mock_record = mock.MagicMock()
        mock_record.raw_value = b'{"value": {"ms": "mocked_ms"}}'

        self.mock_session.handle.fetch = mock.CoroutineMock(return_value=mock_record)

        # Test fetching existing link secret
        secret = await self.holder.get_link_secret()
        assert secret is not None
        mock_load.assert_called_once_with(mock_record.raw_value)
        mock_create.assert_not_called()

    @mock.patch("acapy_agent.indy.credx.holder.LinkSecret.load")
    async def test_get_link_secret_fetch_error(self, mock_load):
        # Mock session to raise an error
        self.mock_session.handle.fetch = mock.CoroutineMock(
            side_effect=AskarError(AskarErrorCode.BACKEND, "Fetch error")
        )

        with pytest.raises(IndyHolderError, match="Error fetching link secret"):
            await self.holder.get_link_secret()

        mock_load.assert_not_called()

    @mock.patch("acapy_agent.indy.credx.holder.LinkSecret.load")
    async def test_get_link_secret_load_error(self, mock_load):
        # Mock session and record
        mock_record = mock.MagicMock()
        mock_record.raw_value = b'{"value": {"ms": "mocked_ms"}}'
        self.mock_session.handle.fetch = mock.CoroutineMock(return_value=mock_record)

        # Mock load to raise an error
        mock_load.side_effect = CredxError(4, "Load error")

        with pytest.raises(IndyHolderError, match="Error loading link secret"):
            await self.holder.get_link_secret()

    @mock.patch("acapy_agent.indy.credx.holder.LinkSecret.load")
    async def test_get_link_secret_fallback_load(self, mock_load):
        # Mock session and record
        mock_record = mock.MagicMock()
        mock_record.raw_value = b'{"value": {"ms": "mocked_ms"}}'
        self.mock_session.handle.fetch = mock.CoroutineMock(return_value=mock_record)

        # Mock load to raise an error initially
        mock_load.side_effect = [CredxError(4, "Load error"), mock.MagicMock()]

        # Test fallback method
        secret = await self.holder.get_link_secret()
        assert secret is not None
        assert mock_load.call_count == 2

    @mock.patch("acapy_agent.indy.credx.holder.LinkSecret.create")
    async def test_get_link_secret_create_error(self, mock_create):
        # Mock session to return no record
        self.mock_session.handle.fetch = mock.CoroutineMock(return_value=None)

        # Mock create to raise an error
        mock_create.side_effect = CredxError(4, "Create error")

        with pytest.raises(IndyHolderError, match="Error creating link secret"):
            await self.holder.get_link_secret()

    @mock.patch("acapy_agent.indy.credx.holder.LinkSecret.create")
    async def test_get_link_secret_create_and_save(self, mock_create):
        # Mock session to return no record
        self.mock_session.handle.fetch = mock.CoroutineMock(return_value=None)

        # Mock successful creation
        mock_secret = mock.MagicMock()
        mock_create.return_value = mock_secret

        # Mock successful insert
        self.mock_session.handle.insert = mock.CoroutineMock()

        # Test creating and saving new link secret
        secret = await self.holder.get_link_secret()
        assert secret is not None
        mock_create.assert_called_once()
        self.mock_session.handle.insert.assert_called_once()

    @mock.patch("acapy_agent.indy.credx.holder.LinkSecret.create")
    async def test_get_link_secret_duplicate_error(self, mock_create):
        # Mock session to return no record
        self.mock_session.handle.fetch = mock.CoroutineMock(return_value=None)

        # Mock successful creation
        mock_secret = mock.MagicMock()
        mock_create.return_value = mock_secret

        # Mock insert to raise a duplicate error
        self.mock_session.handle.insert = mock.CoroutineMock(
            side_effect=[
                AskarError(AskarErrorCode.DUPLICATE, "Duplicate error"),
                mock.CoroutineMock(),
            ]
        )

        # Test handling of duplicate error
        secret = await self.holder.get_link_secret()
        assert secret is not None
        assert mock_create.call_count == 2
        assert self.mock_session.handle.insert.call_count == 2



"""Storage management for configuration-provided mediation invite.

Handle storage and retrieval of mediation invites provided through arguments.
Enables having the mediation invite config be the same
for `provision` and `starting` commands.
"""

import json
from typing import NamedTuple, Optional

from acapy_agent.storage.base import BaseStorage
from acapy_agent.storage.error import StorageNotFoundError
from acapy_agent.storage.record import StorageRecord


class MediationInviteRecord(NamedTuple):
    """A record to store mediation invites and their freshness."""

    invite: str
    used: bool

    def to_json(self) -> str:
        """:return: The current record serialized into a json string."""
        return json.dumps({"invite": self.invite, "used": self.used})

    @staticmethod
    def from_json(json_invite_record: str) -> "MediationInviteRecord":
        """:return: a mediation invite record deserialized from a json string."""
        return MediationInviteRecord(**json.loads(json_invite_record))

    @staticmethod
    def unused(invite: str) -> "MediationInviteRecord":
        """:param invite: invite string as provided by the mediator.

        :return: An unused mediation invitation for the given invite string
        """
        return MediationInviteRecord(invite, False)


class NoDefaultMediationInviteException(Exception):
    """Raised if trying to mark a default invite as used when none exist."""


class MediationInviteStore:
    """Store and retrieve mediation invite configuration."""

    INVITE_RECORD_CATEGORY = "config"
    MEDIATION_INVITE_ID = "mediation_invite"

    def __init__(self, storage: BaseStorage):
        """:param storage: storage facility to be used to store mediation invitation."""
        self.__storage = storage

    async def __retrieve_record(self, key: str) -> Optional[StorageRecord]:
        try:
            return await self.__storage.get_record(self.INVITE_RECORD_CATEGORY, key)
        except StorageNotFoundError:
            return None

    async def store(
        self, mediation_invite: MediationInviteRecord
    ) -> MediationInviteRecord:
        """Store the mediator's invite for further use when starting the agent.

        Update the currently stored invite if one already exists.
        This assumes a new invite and as such, marks it as unused.

        :param mediation_invite: mediation invite url
        :return: stored mediation invite
        """
        current_invite_record = await self.__retrieve_record(self.MEDIATION_INVITE_ID)

        if current_invite_record is None:
            await self.__storage.add_record(
                StorageRecord(
                    type=self.INVITE_RECORD_CATEGORY,
                    id=self.MEDIATION_INVITE_ID,
                    value=mediation_invite.to_json(),
                )
            )
        else:
            await self.__storage.update_record(
                current_invite_record,
                mediation_invite.to_json(),
                tags=current_invite_record.tags,
            )

        return mediation_invite

    async def __retrieve(self) -> Optional[MediationInviteRecord]:
        """:return: the currently stored mediation invite url."""

        invite_record = await self.__retrieve_record(self.MEDIATION_INVITE_ID)
        return (
            MediationInviteRecord.from_json(invite_record.value)
            if invite_record is not None
            else None
        )

    async def __update_mediation_record(
        self, provided_mediation_invitation: str
    ) -> MediationInviteRecord:
        """Update the stored invitation when a new invitation is provided.

        Stored value is only updated if `provided_mediation_invitation` has changed.
        Updated record is marked as unused.

        :param provided_mediation_invitation: mediation invite provided by user
        :return: stored mediation invite
        """
        default_invite = await self.__retrieve()

        if default_invite != provided_mediation_invitation:
            default_invite = await self.store(
                MediationInviteRecord.unused(provided_mediation_invitation)
            )

        return default_invite

    async def mark_default_invite_as_used(self):
        """Mark the currently stored invitation as used if one exists.

        :raises NoDefaultMediationInviteException:
            if trying to mark invite as used when there is no invite stored.
        """
        record = await self.__retrieve()
        if not record:
            raise NoDefaultMediationInviteException(
                "No default mediation invite: cannot mark it as used."
            )

        updated_record = MediationInviteRecord(record.invite, used=True)
        await self.store(updated_record)

        return updated_record

    async def get_mediation_invite_record(
        self, provided_mediation_invitation: Optional[str]
    ) -> Optional[MediationInviteRecord]:
        """Provide the MediationInviteRecord to use/that was used for mediation.

        Returned record may have been used already.

        Stored record is updated if `provided_mediation_invitation` has changed.
        Updated record is marked as unused.

        :param provided_mediation_invitation: mediation invite provided by user
        :return: mediation invite to use/that was used to connect to the mediator. None if
            no invitation was provided/provisioned.
        """

        stored_invite = await self.__retrieve()

        if stored_invite is None and provided_mediation_invitation is None:
            return None
        elif stored_invite is None and provided_mediation_invitation is not None:
            return await self.store(
                MediationInviteRecord.unused(provided_mediation_invitation)
            )
        elif stored_invite is not None and provided_mediation_invitation is None:
            return stored_invite
        elif stored_invite is not None and provided_mediation_invitation is not None:
            return await self.__update_mediation_record(provided_mediation_invitation)



CITIZENSHIP_V1 = {
    "@context": {
        "@version": 1.1,
        "@protected": True,
        "name": "http://schema.org/name",
        "description": "http://schema.org/description",
        "identifier": "http://schema.org/identifier",
        "image": {"@id": "http://schema.org/image", "@type": "@id"},
        "PermanentResidentCard": {
            "@id": "https://w3id.org/citizenship#PermanentResidentCard",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "description": "http://schema.org/description",
                "name": "http://schema.org/name",
                "identifier": "http://schema.org/identifier",
                "image": {"@id": "http://schema.org/image", "@type": "@id"},
            },
        },
        "PermanentResident": {
            "@id": "https://w3id.org/citizenship#PermanentResident",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "ctzn": "https://w3id.org/citizenship#",
                "schema": "http://schema.org/",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                "birthCountry": "ctzn:birthCountry",
                "birthDate": {"@id": "schema:birthDate", "@type": "xsd:dateTime"},
                "commuterClassification": "ctzn:commuterClassification",
                "familyName": "schema:familyName",
                "gender": "schema:gender",
                "givenName": "schema:givenName",
                "lprCategory": "ctzn:lprCategory",
                "lprNumber": "ctzn:lprNumber",
                "residentSince": {"@id": "ctzn:residentSince", "@type": "xsd:dateTime"},
            },
        },
        "Person": "http://schema.org/Person",
    }
}



"""Represents an explicit RFC 15 ack message, adopted into present-proof protocol."""

from typing import Optional

from marshmallow import EXCLUDE, fields, validate

from ....notification.v1_0.messages.ack import V10Ack, V10AckSchema
from ..message_types import PRESENTATION_ACK, PROTOCOL_PACKAGE

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.presentation_ack_handler.PresentationAckHandler"
)


class PresentationAck(V10Ack):
    """Base class representing an explicit ack message for present-proof protocol."""

    class Meta:
        """PresentationAck metadata."""

        handler_class = HANDLER_CLASS
        message_type = PRESENTATION_ACK
        schema_class = "PresentationAckSchema"

    def __init__(
        self,
        status: Optional[str] = None,
        verification_result: Optional[str] = None,
        **kwargs,
    ):
        """Initialize an explicit ack message instance.

        Args:
            status: Status (default OK)
            verification_result: Whether presentation is verified
            kwargs: Additional keyword arguments for message construction

        """
        super().__init__(status, **kwargs)
        self._verification_result = verification_result


class PresentationAckSchema(V10AckSchema):
    """Schema for PresentationAck class."""

    class Meta:
        """PresentationAck schema metadata."""

        model_class = PresentationAck
        unknown = EXCLUDE

    verification_result = fields.Str(
        required=False,
        validate=validate.OneOf(["true", "false"]),
        metadata={
            "description": "Whether presentation is verified: true or false",
            "example": "true",
        },
    )



"""Base Proof Purpose class."""

from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Optional

from ....messaging.util import str_to_datetime
from ..document_loader import DocumentLoaderMethod
from ..validation_result import PurposeResult

# Avoid circular dependency
if TYPE_CHECKING:
    from ..suites import LinkedDataProof


class ProofPurpose:
    """Base proof purpose class."""

    def __init__(
        self,
        *,
        term: str,
        date: Optional[datetime] = None,
        max_timestamp_delta: Optional[timedelta] = None,
    ):
        """Initialize new proof purpose instance."""
        self.term = term
        self.date = date or datetime.now()
        self.max_timestamp_delta = max_timestamp_delta

    def validate(
        self,
        *,
        proof: dict,
        document: dict,
        suite: "LinkedDataProof",
        verification_method: dict,
        document_loader: DocumentLoaderMethod,
    ) -> PurposeResult:
        """Validate whether created date of proof is out of max_timestamp_delta range."""
        try:
            if self.max_timestamp_delta is not None:
                expected = self.date.timestamp()

                created = str_to_datetime(proof.get("created")).timestamp()

                if not (
                    created >= (expected - self.max_timestamp_delta.total_seconds())
                    and created <= (expected + self.max_timestamp_delta.total_seconds())
                ):
                    raise Exception("The proof's created timestamp is out of range.")

            return PurposeResult(valid=True)
        except Exception as err:
            return PurposeResult(valid=False, error=str(err))

    def update(self, proof: dict) -> dict:
        """Update proof purpose on proof."""
        proof["proofPurpose"] = self.term
        return proof

    def match(self, proof: dict) -> bool:
        """Check whether the passed proof matches with the term of this proof purpose."""
        return proof.get("proofPurpose") == self.term

    def __eq__(self, o: object) -> bool:
        """Check if object is same as ProofPurpose."""
        if isinstance(o, ProofPurpose):
            return (
                self.date == o.date
                and self.term == o.term
                and self.max_timestamp_delta == o.max_timestamp_delta
            )

        return False



"""Ledger Request Executor."""

from typing import Optional, Tuple

from ...config.base import InjectionError
from ...core.profile import Profile
from ...ledger.base import BaseLedger
from ...ledger.multiple_ledger.base_manager import (
    BaseMultipleLedgerManager,
    MultipleLedgerManagerError,
)

(
    GET_SCHEMA,
    GET_CRED_DEF,
    GET_REVOC_REG_DEF,
    GET_REVOC_REG_ENTRY,
    GET_KEY_FOR_DID,
    GET_ALL_ENDPOINTS_FOR_DID,
    GET_ENDPOINT_FOR_DID,
    GET_NYM_ROLE,
    GET_REVOC_REG_DELTA,
) = tuple(range(9))


class IndyLedgerRequestsExecutor:
    """Executes Ledger Requests based on multiple ledger config, if set."""

    def __init__(
        self,
        profile: Profile,
    ):
        """Initialize IndyLedgerRequestsExecutor.

        Args:
            profile: The active profile instance

        """
        self.profile = profile

    async def get_ledger_for_identifier(
        self, identifier: str, txn_record_type: int
    ) -> Tuple[Optional[str], Optional[BaseLedger]]:
        """Return ledger info given the record identifier."""
        # For seqNo
        if identifier.isdigit():
            return (None, self.profile.inject(BaseLedger))
        elif (
            self.profile.settings.get("ledger.ledger_config_list")
            and len(self.profile.settings.get("ledger.ledger_config_list")) > 0
        ):
            try:
                multiledger_mgr = self.profile.inject(BaseMultipleLedgerManager)
                extracted_did = multiledger_mgr.extract_did_from_identifier(identifier)
                if txn_record_type in tuple(range(4)):
                    cache_did = True
                else:
                    cache_did = False
                return await multiledger_mgr.lookup_did_in_configured_ledgers(
                    extracted_did, cache_did=cache_did
                )
            except (MultipleLedgerManagerError, InjectionError):
                pass
        return (None, self.profile.inject_or(BaseLedger))

    async def get_ledger_inst(self, ledger_id: str) -> Optional[BaseLedger]:
        """Return ledger instance from ledger_id set in config."""
        multiledger_mgr = self.profile.inject(BaseMultipleLedgerManager)
        return await multiledger_mgr.get_ledger_inst_by_id(ledger_id=ledger_id)



from .bbs_v1 import BBS_V1
from .citizenship_v1 import CITIZENSHIP_V1
from .credentials_v1 import CREDENTIALS_V1
from .credentials_v2 import CREDENTIALS_V2
from .did_v1 import DID_V1
from .dif_presentation_submission_v1 import DIF_PRESENTATION_SUBMISSION_V1
from .ed25519_2020_v1 import ED25519_2020_V1
from .examples_v1 import EXAMPLES_V1
from .multikey_v1 import MULTIKEY_V1
from .odrl import ODRL
from .schema_org import SCHEMA_ORG
from .security_v1 import SECURITY_V1
from .security_v2 import SECURITY_V2
from .security_v3_unstable import SECURITY_V3_UNSTABLE
from .vaccination_v1 import VACCINATION_V1

__all__ = [
    "DID_V1",
    "DIF_PRESENTATION_SUBMISSION_V1",
    "SECURITY_V1",
    "SECURITY_V2",
    "SECURITY_V3_UNSTABLE",
    "BBS_V1",
    "ED25519_2020_V1",
    "MULTIKEY_V1",
    "CREDENTIALS_V1",
    "CREDENTIALS_V2",
    "CITIZENSHIP_V1",
    "VACCINATION_V1",
    "EXAMPLES_V1",
    "ODRL",
    "SCHEMA_ORG",
]



"""AnonCreds credential definition proposal."""

import re

from marshmallow import fields, validate

from ...core.profile import Profile
from ...messaging.models.openapi import OpenAPISchema
from ...messaging.valid import (
    ANONCREDS_CRED_DEF_ID_EXAMPLE,
    ANONCREDS_CRED_DEF_ID_VALIDATE,
    ANONCREDS_DID_EXAMPLE,
    ANONCREDS_DID_VALIDATE,
    ANONCREDS_SCHEMA_ID_EXAMPLE,
    ANONCREDS_SCHEMA_ID_VALIDATE,
    INDY_DID_VALIDATE,
    MAJOR_MINOR_VERSION_EXAMPLE,
    MAJOR_MINOR_VERSION_VALIDATE,
)


class AnonCredsCredentialDefinitionProposal(OpenAPISchema):
    """Query string parameters for credential definition searches."""

    cred_def_id = fields.Str(
        required=False,
        validate=ANONCREDS_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition id. This is the only required field.",
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
    )
    issuer_id = fields.Str(
        required=False,
        # TODO: INDY_DID_VALIDATE should be removed when indy sov did's
        # are represented by did:sov:{nym} in acapy
        validate=validate.NoneOf([ANONCREDS_DID_VALIDATE, INDY_DID_VALIDATE]),
        metadata={"description": "Issuer DID", "example": ANONCREDS_DID_EXAMPLE},
    )
    schema_id = fields.Str(
        required=False,
        validate=ANONCREDS_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        },
    )
    schema_issuer_id = fields.Str(
        required=False,
        # TODO: INDY_DID_VALIDATE should be removed when indy sov did's
        # are represented by did:sov:{nym} in acapy
        validate=validate.NoneOf([ANONCREDS_DID_VALIDATE, INDY_DID_VALIDATE]),
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        },
    )
    schema_name = fields.Str(
        required=False, metadata={"description": "Schema name", "example": "simple"}
    )
    schema_version = fields.Str(
        required=False,
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Schema version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )


CRED_DEF_TAGS = list(
    vars(AnonCredsCredentialDefinitionProposal).get("_declared_fields", [])
)

CRED_DEF_EVENT_PREFIX = "acapy::CRED_DEF::"
EVENT_LISTENER_PATTERN = re.compile(f"^{CRED_DEF_EVENT_PREFIX}(.*)?$")


async def notify_cred_def_event(
    profile: Profile, cred_def_id: str, meta_data: dict
) -> None:
    """Send notification for a cred def post-process event."""
    await profile.notify(
        CRED_DEF_EVENT_PREFIX + cred_def_id,
        meta_data,
    )



"""Test PeerDIDResolver."""

import pytest
import pytest_asyncio

from ....core.profile import Profile
from ....utils.testing import create_test_profile
from ..peer2 import PeerDID2Resolver

TEST_DID0 = "did:peer:2.Ez6LSpkcni2KTTxf4nAp6cPxjRbu26Tj4b957BgHcknVeNFEj.Vz6MksXhfmxm2i3RnoHH2mKQcx7EY4tToJR9JziUs6bp8a6FM.SeyJ0IjoiZGlkLWNvbW11bmljYXRpb24iLCJzIjoiaHR0cDovL2hvc3QuZG9ja2VyLmludGVybmFsOjkwNzAiLCJyZWNpcGllbnRfa2V5cyI6W119"


@pytest.fixture
def resolver():
    """Resolver fixture."""
    yield PeerDID2Resolver()


@pytest_asyncio.fixture
async def profile():
    """Profile fixture."""
    profile = await create_test_profile()
    yield profile


@pytest.mark.asyncio
async def test_resolver(profile: Profile, resolver: PeerDID2Resolver):
    """Test resolver setup."""
    assert resolver.supported_did_regex
    assert await resolver.supports(profile, TEST_DID0)
    doc = await resolver.resolve(profile, TEST_DID0)
    assert doc
    assert doc["id"] == TEST_DID0
    assert "service" in doc
    assert "verificationMethod" in doc
    assert len(doc["verificationMethod"]) == 2
